# Implementation based on tf.keras.engine.data_adapter.py
# https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/keras/engine/data_adapter.py


import math
import typing as tp
from operator import itemgetter

import jax.numpy as jnp
import numpy as np

from .data_adapter import DataAdapter
from .utils import flatten, map_structure, pack_x_y_sample_weight

DEFAULT_BATCH_SIZE = 32


class ArrayDataAdapter(DataAdapter):
    """Adapter that handles NumPy and Jax numpy arrays."""

    @staticmethod
    def can_handle(x, y=None):
        flat_inputs = list(flatten(x))
        if y is not None:
            flat_inputs += list(flatten(y))

        supported_types = (jnp.ndarray, np.ndarray)
        # if pd:
        #     supported_types = (ops.Tensor, np.ndarray, pd.Series, pd.DataFrame)

        def _is_array(v):
            if isinstance(v, supported_types):
                return True
            return False

        return all(_is_array(v) for v in flat_inputs)

    def __init__(
        self,
        x,
        y=None,
        sample_weights: tp.Union[jnp.ndarray, np.ndarray, None] = None,
        batch_size: tp.Optional[int] = None,
        epochs: int = 1,
        steps: tp.Optional[int] = None,
        shuffle: bool = False,
        drop_remainder: bool = False,
        **kwargs,
    ):
        super(ArrayDataAdapter, self).__init__(x, y, **kwargs)

        inputs = pack_x_y_sample_weight(x, y, sample_weights)

        num_samples = set(int(i.shape[0]) for i in flatten(inputs))

        if len(num_samples) > 1:
            msg = "Data cardinality is ambiguous:\n"
            for label, data in zip(["x", "y", "sample_weight"], inputs):
                msg += "  {} sizes: {}\n".format(
                    label, ", ".join(str(i.shape[0]) for i in data)
                )
            msg += "Please provide data which shares the same first dimension."
            raise ValueError(msg)

        num_samples = (
            num_samples.pop()
            if num_samples
            else batch_size
            if batch_size is not None
            else DEFAULT_BATCH_SIZE
        )

        # If batch_size is not passed but steps is, calculate from the input data.
        if batch_size is None:
            # if batch_size is None and steps is None:
            #     raise ValueError("Please provide either batch_size or steps")
            batch_size = (
                int(math.ceil(num_samples / steps)) if steps else DEFAULT_BATCH_SIZE
            )

        self._size = int(math.ceil(num_samples / batch_size))
        self._batch_size = batch_size

        num_full_batches = int(num_samples // batch_size)
        self._partial_batch_size = num_samples % batch_size

        self._shuffle = shuffle

        dataset_indices = np.arange(num_samples)

        def dataset_generator():
            while True:
                if shuffle:
                    np.random.shuffle(dataset_indices)

                for batch in range(
                    num_full_batches + int(self._partial_batch_size != 0)
                ):
                    indices = dataset_indices[
                        batch * batch_size : (batch + 1) * batch_size
                    ]

                    # # Drop last batch
                    # if drop_remainder and len(indices) < batch_size:
                    #     print("Dropping!")
                    #     continue
                    inputs_slices = map_structure(itemgetter(indices), inputs)

                    yield inputs_slices

        self._dataset = dataset_generator

    def get_dataset(self):
        return self._dataset

    def get_size(self):
        return self._size

    @property
    def batch_size(self):
        return self._batch_size

    def has_partial_batch(self):
        return self._partial_batch_size > 0

    @property
    def partial_batch_size(self):
        return self._partial_batch_size or None

    def should_recreate_iterator(self):
        # An infinite dataset is always created here.
        return False



from functools import partial
from typing import List, Optional, Sequence, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..ops import locations_to_labels
from .common import ChannelAttention


class LPN(nn.Module):
    """
    Args:
    in_channels: number of channels in input feature
    feature_level: input feature level default 3
    conv_layers: conv layer spec
    with_channel_attention: whether include channel attention
    """

    feature_levels: Sequence[int] = (4, 3, 2)
    conv_spec: Tuple[Sequence[int], Sequence[int]] = ((256, 256, 256, 256), ())
    detection_roi: float = 8.0

    @nn.compact
    def _process_feature(self, feature: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        conv_spec = self.conv_spec

        x = feature

        for n_ch in conv_spec[0]:
            x = nn.Conv(n_ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=n_ch)(x[None, ...])[0]
            x = jax.nn.relu(x)

        for n_ch in conv_spec[1]:
            x = nn.Conv(n_ch, (1, 1), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=n_ch)(x[None, ...])[0]
            x = jax.nn.relu(x)

        scores_out = nn.Conv(1, (1, 1))(x)
        scores_out = jax.nn.sigmoid(scores_out)

        regression_out = nn.Conv(2, (1, 1))(x)

        return scores_out, regression_out

    def __call__(
        self, inputs: dict, scaled_gt_locations: jnp.ndarray = None, *, training=None
    ) -> dict:
        """
        Args:
            inputs: feature dict: {'lvl': [H, W, C]}
            scaled_gt_locations: scaled 0..1 [N, 2], only valid in training
        Returns:
            outputs:
            {
                lpn_scores: dict: {'lvl': [H, W, 1]}
                lpn_regressions: dict {'lvl': [H, W, 2]}
                gt_lpn_scores: dict {'lvl': [H, W, 1]}, only if training
                gt_lpn_regressions: dict {'lvl': [H, W, 2]}, only if training
            }
        """

        all_scores = dict()
        all_regrs = dict()
        all_gt_scores = dict()
        all_gt_regrs = dict()

        for lvl in self.feature_levels:

            feature = inputs[str(lvl)]
            score, regression = self._process_feature(feature)
            all_scores[str(lvl)] = score
            all_regrs[str(lvl)] = regression

            if training and scaled_gt_locations is not None:
                score_target, regression_target = locations_to_labels(
                    scaled_gt_locations,
                    target_shape=feature.shape[-3:-1],
                    threshold=self.detection_roi / (2**lvl),
                )
                all_gt_scores[str(lvl)] = score_target
                all_gt_regrs[str(lvl)] = regression_target

        outputs = dict(
            lpn_scores=all_scores,
            lpn_regressions=all_regrs,
            lpn_gt_scores=all_gt_scores,
            lpn_gt_regressions=all_gt_regrs,
        )

        return outputs



from functools import partial

import jax
import jax.numpy as jnp
import optax

from ..ops import sorbel_edges
from ..train.loss import Loss
from .detection import _binary_focal_crossentropy

EPS = jnp.finfo("float32").eps


def _compute_edge(instance_output, instance_yc, instance_xc, height, width):
    _, ps, _ = instance_yc.shape
    padding = ps // 2 + 1

    patch_edges = jnp.square(sorbel_edges(instance_output))
    patch_edges = (patch_edges[0] + patch_edges[1]) / 8.0
    patch_edges = jnp.sqrt(jnp.clip(patch_edges, 1e-8, 1.0))  # avoid GPU error
    # patch_edges = jnp.where(patch_edges > 0, jnp.sqrt(patch_edges), 0)
    combined_edges = jnp.zeros([height + padding * 2, width + padding * 2])
    combined_edges = combined_edges.at[
        instance_yc + padding, instance_xc + padding
    ].add(patch_edges)
    combined_edges = combined_edges[
        padding : padding + height, padding : padding + width
    ]
    combined_edges = jnp.tanh(combined_edges)

    return combined_edges


def self_supervised_edge_losses(pred, input):
    """
    Args:
        auxnet_out: [H,W, n_groups]
        instance_edge: [H,W]
        category: category number [0, n_groups)
    return:
        loss: float
    """
    instance_output = pred["instance_output"]
    instance_yc = pred["instance_yc"]
    instance_xc = pred["instance_xc"]
    height, width, _ = input["image"].shape
    instance_edge = _compute_edge(
        instance_output, instance_yc, instance_xc, height, width
    )
    instance_edge_pred = jax.nn.sigmoid(pred["edge_pred"])

    return optax.l2_loss(instance_edge_pred, instance_edge).mean()


# def auxnet_losses(auxnet_out, instance_edge, category=None):
#     '''
#     Args:
#         auxnet_out: [H,W, n_groups]
#         instance_edge: [H,W]
#         category: category number [0, n_groups)
#     return:
#         loss: float
#     '''
#     edge_pred, instance_edge = _get_auxnet_prediction(auxnet_out, instance_edge, category)

#     return _binary_focal_crossentropy(edge_pred, instance_edge >= 0.5).mean()

# class AuxnetLoss(Loss):
#     def call(
#         self,
#         preds: dict,
#         group_num: jnp.ndarray = None,
#     ):
#         if not 'training_locations' in preds:
#             return 0.0

#         return jax.vmap(auxnet_losses)(
#             auxnet_out = preds['auxnet_out'],
#             instance_edge = preds['instance_edge'],
#             category = group_num.astype(int),
#         )


class AuxEdgeLoss(Loss):
    def call(
        self,
        inputs: dict,
        preds: dict,
        **kwargs,
    ):
        if not "training_locations" in preds:
            return 0.0

        return self_supervised_edge_losses(preds, inputs)


class AuxSizeLoss(Loss):
    def __init__(self, alpha):
        super().__init__()
        self.a = alpha

    def call(self, inputs, preds, **kwargs):
        height, width, _ = inputs["image"].shape
        valid_locs = (
            (preds["instance_yc"] >= 0)
            & (preds["instance_yc"] < height)
            & (preds["instance_xc"] >= 0)
            & (preds["instance_xc"] < width)
        )
        _, crop_size, _ = preds["instance_output"].shape
        areas = (
            jnp.sum(
                preds["instance_output"], axis=(1, 2), where=valid_locs, keepdims=True
            )
            / crop_size
            / crop_size
        )
        areas = jnp.clip(areas, EPS, 1.0)

        loss = self.a * jnp.clip(jax.lax.rsqrt(areas), 0.0, 1.0e8)
        loss = jnp.sum(loss, axis=0, where=preds["instance_mask"]) / (
            jnp.count_nonzero(preds["instance_mask"]) + EPS
        )

        return loss


class AuxSegLoss(Loss):
    def call(self, inputs, preds, **kwargs):
        height, width, _ = inputs["image"].shape
        _, ps, _ = preds["instance_yc"].shape

        # def _to_patch(img, yc, xc):
        #     _, ps, _ = yc.shape
        #     padding = [[ps // 2, ps // 2], [ps // 2, ps // 2]]
        #     img = jnp.pad(img, padding, constant_values=-1.0)
        #     return img[yc + ps // 2, xc + ps // 2]

        def _max_merge(pred):
            label = jnp.zeros([height + ps, width + ps]) - 1.0e8
            yc, xc = pred["instance_yc"], pred["instance_xc"]
            label = label.at[yc + ps // 2, xc + ps // 2].max(pred["instance_logit"])
            label = label[ps // 2 : -ps // 2, ps // 2 : -ps // 2]
            return label

        fg_pred = jax.nn.tanh(preds["fg_pred"])
        fg = jax.nn.tanh(_max_merge(preds))
        loss = 1.0 - fg_pred * fg
        loss = loss.mean(axis=(1, 2))

        return loss



from functools import partial
from typing import List, Optional, Sequence, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..ops import location_matching, sorted_non_max_suppression


class Detector(nn.Module):
    """A weightless layer that conver LPN output to a list of locations"""

    train_nms_threshold: float = 8.0
    train_pre_nms_topk: int = -1
    train_max_output: int = 768
    train_min_score: float = 0.2
    max_proposal_offset: float = 12
    test_nms_threshold: float = 8.0
    test_pre_nms_topk: int = -1
    test_max_output: int = 768
    test_min_score: float = 0.25
    training: bool = True

    def _proposal_locations(self, lpn_scores, lpn_regression, training: bool = None):
        """
        Produce a list of proposal locations based on predication map, remove redundency with non_max_suppression
        Args:
            lpn_score: dict {'lvl': [height, width, 1]} lpn score prediction
            lpn_regression: dict {'lvl': [height, width, 2]} lpn regression prediction
        Returns:
            scores: [N], sorted scores
            locations: [N, 2], proposed location
        """
        if training:
            distance_threshold = self.train_nms_threshold
            output_size = self.train_max_output
            topk = self.train_pre_nms_topk
            score_threshold = self.train_min_score
        else:
            distance_threshold = self.test_nms_threshold
            output_size = self.test_max_output
            topk = self.test_pre_nms_topk
            score_threshold = self.test_min_score

        # preprocess
        scores = []
        locations = []
        for k in lpn_scores:
            s = lpn_scores[k].squeeze(-1)
            r = lpn_regression[k]
            height, width = s.shape
            loc = jnp.mgrid[:height, :width] + 0.5
            loc = loc.transpose(1, 2, 0) + r

            # set invalid location score to 0
            is_valid = (
                (loc > 0.0).all(axis=-1)
                & (loc[:, :, 0] < height)
                & (loc[:, :, 1] < width)
            )
            s = jnp.where(is_valid, s, 0)

            loc *= 2 ** int(k)  # scale to origianl image size

            locations.append(loc.reshape(-1, 2))
            scores.append(s.reshape(-1))

        scores = jnp.concatenate(scores, axis=0)
        locations = jnp.concatenate(locations, axis=0)

        # select topk
        if topk <= 0 or topk > scores.size:
            topk = scores.size
        scores, selections = jax.lax.top_k(scores, topk)
        locations = locations[selections]

        # refine
        if distance_threshold > 0:  # nms
            threshold = 1 / distance_threshold / distance_threshold
            scores, locations = sorted_non_max_suppression(
                scores,
                locations,
                output_size,
                threshold,
                score_threshold,
            )
        else:
            is_valid = scores >= score_threshold
            scores = jnp.where(is_valid, scores, -1.0)
            locations = jnp.where(is_valid[:, None], locations, -1.0)

        return scores, locations

    def _gen_train_locations(self, gt_locations, pred_locations):
        """replacing gt_locations with pred_locations if the close enough
        gt_locations: [N, 2] tensor
        pred_locations: [M, 2] tensor, sorted with scores

        1. Each pred_location is matched to the closest gt_location
        2. For each gt_location, pick the matched pred_location with highest score
        3. if the picked pred_location is within threshold distance, replace the gt_location with the pred_location
        """

        threshold = self.max_proposal_offset

        n_gt_locs = gt_locations.shape[0]
        n_pred_locs = pred_locations.shape[0]

        matched_id, indicators = location_matching(
            pred_locations, gt_locations, threshold
        )
        matched_id = jnp.where(indicators, matched_id, -1)

        matching_matrix = (
            matched_id[None, :] == jnp.arange(n_gt_locs)[:, None]
        )  # true at matched gt(row)/pred(col), at most one true per col
        last_col = jnp.ones([n_gt_locs, 1], dtype=bool)
        matching_matrix = jnp.concatenate(
            [matching_matrix, last_col], axis=-1
        )  # last col is true

        # first true of every row
        matched_loc_ids = jnp.argmax(matching_matrix, axis=-1)

        training_locations = jnp.where(
            matched_loc_ids[:, None] == n_pred_locs,  # true: failed match
            gt_locations,
            pred_locations[
                matched_loc_ids, :
            ],  # out-of-bound error silently dropped in jax
        )

        return training_locations

    def __call__(
        self,
        scores: jnp.ndarray,
        regressions: jnp.ndarray,
        gt_locations: jnp.ndarray = None,
        *,
        training: bool = None,
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Args:
            score: {'lvl', [H, W, 1]) output from LPN
            regression: ('lvl': [H, W, 2]) output from LPN
            gt_locations: [N, 2] during training or None during inference. Maybe padded with -1
        Outputs:
            dict {
                training_locations: [N, 2] only during training
                pred_locations: [M, 2] sorted based on scores; M == test_max_output; padded with -1
                pred_scores: [M] sorted, padded with -1
            }
        """

        scores = jax.lax.stop_gradient(scores)
        regressions = jax.lax.stop_gradient(regressions)

        if training and self.max_proposal_offset <= 0:
            return dict(
                training_locations=gt_locations,
            )
        proposed_scores, proposed_locations = self._proposal_locations(
            scores, regressions, training
        )
        outputs = dict(
            pred_locations=proposed_locations,
            pred_scores=proposed_scores,
        )
        if training:
            outputs.update(
                dict(
                    training_locations=self._gen_train_locations(
                        gt_locations,
                        proposed_locations,
                    )
                )
            )
        return outputs



from typing import Callable

import jax

from .boxes import box_iou_similarity
from .locations import distance_similarity

jnp = jax.numpy

NMS_TILE_SIZE = 512


def _suppress(boxes, mask):
    return jnp.where(mask[:, None], -1, boxes)


def _suppression_loop_body(inputs):
    """Process boxes in the range [idx*NMS_TILE_SIZE, (idx+1)*NMS_TILE_SIZE).
    Args:
        inputs: tuple
            idx: current slice
            boxes: a tensor with a shape of [N, 4].
            num_selected: number of selected boxes so far
            threshold: float
    Returns:
        boxes: updated
        num_selected: updated
    """

    # Iterates over tiles that can possibly suppress the current tile.

    idx, boxes, num_selected, threshold = inputs

    box_slice = boxes[idx]
    similarity_func = (
        box_iou_similarity if boxes.shape[-1] == 4 else distance_similarity
    )

    def _for_loop_func(idx, slice):
        iou = similarity_func(boxes[idx], slice)
        suppressed = jnp.any(iou >= threshold, axis=0)
        return _suppress(slice, suppressed)

    box_slice = jax.lax.fori_loop(0, idx, _for_loop_func, box_slice)

    # Iterates over the current tile to compute self-suppression.
    iou = similarity_func(box_slice, box_slice)
    mask = jnp.arange(NMS_TILE_SIZE).reshape([1, -1]) > jnp.arange(
        NMS_TILE_SIZE
    ).reshape([-1, 1])
    mask = mask & (iou > threshold)

    def _while_loop_func(inputs):
        mask, cnt = inputs
        cnt = jnp.count_nonzero(mask)
        can_suppress_others = ~mask.any(axis=0)
        suppressed = (mask & can_suppress_others[:, None]).any(axis=0)
        mask = mask & ~suppressed[:, None]
        # suppressed = mask.at[can_suppress_others, :].any(axis=0)
        # mask = mask.at[suppressed, :].set(False)
        return mask, cnt

    def _while_cond_func(inputs):
        mask, cnt = inputs
        return jnp.count_nonzero(mask) != cnt

    mask, _ = jax.lax.while_loop(_while_cond_func, _while_loop_func, (mask, 0))
    box_slice = _suppress(box_slice, mask.any(axis=0))

    # Uses box_slice to update the input boxes.
    boxes = boxes.at[idx].set(box_slice)

    # output_size.
    num_selected += jnp.count_nonzero((box_slice >= 0).any(axis=-1))

    return boxes, num_selected


def sorted_non_max_suppression(
    scores,
    boxes,
    max_output_size: int,
    threshold: float = 0.5,
    min_score: float = 0,
):
    """non-maximum suppression.
    Assumption:
        * The boxes are sorted by scores
    The overal design of the algorithm is to handle boxes tile-by-tile:
    Args:
        scores: [N]
        boxes: [N, C]  C=4 for boxes, C=2 for locations
        max_output_size: a positive scalar integer
        threshold: a scalar float, can be negative
        min_score: min score to be selected, default 0
    Returns:
        nms_scores: [M].  M = max_output_size
        nms_proposals: [M, C].
    """
    # preprocessing
    c = boxes.shape[-1]
    if c != 2 and c != 4:
        raise ValueError(f"boxes should be Nx4 or Nx2, got Nx{c}")
    # if similarity_func is None:
    #     similarity_func = box_iou_similarity if c == 4 else distance_similarity

    if max_output_size <= 0:
        max_output_size = boxes.shape[0]

    # pad_to_multiply_of(tile_size)
    num_boxes = boxes.shape[0]
    pad = NMS_TILE_SIZE - 1 - (num_boxes - 1) % NMS_TILE_SIZE
    boxes = jnp.pad(boxes, [[0, pad], [0, 0]], constant_values=-1)
    scores = jnp.pad(scores, [[0, pad]], constant_values=-1)
    num_boxes += pad
    boxes = boxes.reshape(-1, NMS_TILE_SIZE, c)

    def _trivial_suppress_all(inputs):
        (
            idx,
            boxes,
            num_outputs,
            _,
        ) = inputs
        boxes = boxes.at[idx].set(-1)
        return boxes, num_outputs

    # process all tiles until generating enough output
    num_selected = 0
    for idx in range(num_boxes // NMS_TILE_SIZE):
        boxes, num_selected = jax.lax.cond(
            (scores[idx * NMS_TILE_SIZE] >= min_score)
            & (num_selected < max_output_size),
            _suppression_loop_body,
            _trivial_suppress_all,
            (idx, boxes, num_selected, threshold),
        )

    # reshape boxes back
    boxes = boxes.reshape(-1, c)

    # remove suppressed boxes
    selected = jnp.argwhere(
        (boxes >= 0).any(axis=-1), size=max_output_size, fill_value=-1
    ).squeeze(-1)
    scores = jnp.where(selected >= 0, scores[selected], -1.0)
    boxes = jnp.where(selected[:, None] >= 0, boxes[selected, :], -1.0)

    # remove low score
    selected = scores >= min_score
    scores = jnp.where(selected, scores, -1)
    boxes = jnp.where(selected[:, None], boxes, -1)

    return scores, boxes



import glob
import json

import imageio
import numpy as np
import tensorflow as tf

try:
    from pycocotools.coco import COCO
except:
    pass
from os.path import join

from skimage.measure import regionprops


# FIXME avoid dependency on COCO
def _coco_generator(annotation_file, image_path):
    coco = COCO(annotation_file=annotation_file)
    for imgid in coco.getImgIds():
        imginfo = coco.imgs[imgid]
        bboxes = []
        mis = []
        locs = []
        rls = []
        for ann_id in coco.getAnnIds(imgIds=imgid):
            ann = coco.anns[ann_id]
            bbox = ann["bbox"]
            bbox = np.array([bbox[1], bbox[0], bbox[1] + bbox[3], bbox[0] + bbox[2]])
            bboxes.append(bbox)

            mask = coco.annToMask(ann)
            mi = np.stack(np.where(mask >= 0.5), axis=-1)
            mis.append(mi)
            rls.append(mi.shape[0])

            locs.append(mi.mean(axis=0))

        bboxes = np.array(bboxes, dtype="float32")
        locs = np.array(locs, dtype="float32")
        mis = np.concatenate(mis, dtype="int64")
        rls = np.array(rls, dtype="int64")

        filepath = glob.glob(join(image_path, "**", imginfo["file_name"]))
        assert len(filepath) == 1

        image = imageio.imread(filepath[0])
        if len(image.shape) == 2:
            image = image[:, :, None]
        image = (image / 255.0).astype("float32")

        binary_mask = np.zeros(image.shape[:2], "float32")
        binary_mask[tuple(mis.transpose())] = 1.0

        mis = tf.RaggedTensor.from_row_lengths(mis, rls)

        yield {
            "img_id": imgid,
            "image": image,
            "mask_indices": mis,
            "locations": locs,
            "binary_mask": binary_mask,
            "bboxes": bboxes,
            "filename": imginfo["file_name"],
        }


def dataset_from_coco_annotations(
    annotation_file, image_path, image_shape=[None, None, 3]
):
    return tf.data.Dataset.from_generator(
        lambda: _coco_generator(annotation_file, image_path),
        output_signature={
            "img_id": tf.TensorSpec([], dtype=tf.int64),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            "mask_indices": tf.RaggedTensorSpec([None, None, 2], tf.int64, 1),
            "locations": tf.TensorSpec([None, 2], dtype=tf.float32),
            "binary_mask": tf.TensorSpec([None, None], dtype=tf.float32),
            "bboxes": tf.TensorSpec([None, 4], dtype=tf.float32),
            "filename": tf.TensorSpec([], dtype=tf.string),
        },
    )


def _simple_generator(annotation_file, image_path, normalize=True):
    with open(annotation_file, "r") as f:
        annotations = json.load(f)
    for ann in annotations:
        image = imageio.imread(join(image_path, ann["image_file"]))
        if len(image.shape) == 2:
            image = image[:, :, None]
        # image = (image / 255.0).astype('float32')
        image = image.astype("float32")
        if normalize:
            image -= image.mean()
            image /= image.std()

        binary_mask = imageio.imread(join(image_path, ann["mask_file"]))
        if len(binary_mask.shape) == 3:
            binary_mask = binary_mask[:, :, 0]
        binary_mask = (binary_mask > 0).astype("float32")

        locations = np.array(ann["locations"]).astype("float32")

        yield {
            "img_id": ann["img_id"],
            "image": image,
            "binary_mask": binary_mask,
            "locations": locations,
        }


def dataset_from_simple_annotations(
    annotation_file, image_path, image_shape=[None, None, 3], **kwargs
):
    return tf.data.Dataset.from_generator(
        lambda: _simple_generator(annotation_file, image_path, **kwargs),
        output_signature={
            "img_id": tf.TensorSpec([], dtype=tf.int64),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            "locations": tf.TensorSpec([None, 2], dtype=tf.float32),
            "binary_mask": tf.TensorSpec([None, None], dtype=tf.float32),
        },
    )


def _img_mask_pair_generator(ds_files, normalize=True):
    for k, (img_file, mask_file) in enumerate(ds_files):
        img = imageio.imread(img_file).astype("float32")
        if normalize:
            img -= img.mean()
            img /= img.std()
        if len(img.shape) == 2:
            img = img[:, :, None]
        mask = imageio.imread(mask_file)
        if len(mask.shape) == 3:
            mask = mask[:, :, 0]
        binary_mask = (mask > 0).astype("float32")

        bboxes = []
        locs = []
        # mis = []
        # mi_lengths = []
        for prop in regionprops(mask):
            bboxes.append(prop["bbox"])
            locs.append(prop["centroid"])
            # mi = np.array(np.where(prop['image'])).transpose()
            # mi = mi + bboxes[-1][:2]
            # mi_lengths.append(mi.shape[0])
            # mis.append(mi)
        bboxes = np.array(bboxes, dtype="float32")
        locs = np.array(locs, dtype="float32")
        # mis = tf.RaggedTensor.from_row_lengths(np.concatenate(mis), mi_lengths)
        yield {
            "img_id": k,
            "image": img,
            "locations": locs,
            "binary_mask": binary_mask,
            "bboxes": bboxes,
            "label": mask,
            # 'mask_indices': mis
        }


def dataset_from_img_mask_pairs(
    imgfiles, maskfiles, image_shape=[None, None, 3], **kwargs
):
    return tf.data.Dataset.from_generator(
        lambda: _img_mask_pair_generator(zip(imgfiles, maskfiles), **kwargs),
        output_signature={
            "img_id": tf.TensorSpec([], dtype=tf.int64),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            # 'mask_indices': tf.RaggedTensorSpec([None, None, 2], tf.int64, 1),
            "locations": tf.TensorSpec([None, 2], dtype=tf.float32),
            "binary_mask": tf.TensorSpec([None, None], dtype=tf.float32),
            "bboxes": tf.TensorSpec([None, 4], dtype=tf.float32),
            "label": tf.TensorSpec([None, None], dtype=tf.int32),
        },
    )



import typing as tp
from dataclasses import asdict, field
from functools import partial

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..ops import *
from .auxiliary import *
from .convnext import ConvNeXt
from .detector import Detector
from .lpn import LPN
from .resnet import ResNet
from .segmentor import Segmentor


class Lacss(nn.Module):
    backbone: type = "ResNet"
    backbone_cfg: dict = field(default_factory=dict)
    lpn: dict = field(default_factory=dict)
    detector: dict = field(default_factory=dict)
    segmentor: dict = field(default_factory=dict)

    def setup(self):
        cls = globals()[self.backbone]
        self._backbone = cls(**self.backbone_cfg)
        self._lpn = LPN(**self.lpn)
        self._detector = Detector(**self.detector)
        self._segmentor = Segmentor(**self.segmentor)

    def __call__(
        self,
        image: jnp.ndarray,
        gt_locations: jnp.ndarray = None,
        *,
        training: bool = None,
    ) -> dict:
        """
        Args:
            image: [H, W, C]
            gt_locations: [M, 2] if training, otherwise None
        Returns:
            a dict of model outputs
        """
        orig_height, orig_width, ch = image.shape
        if ch == 1:
            image = jnp.repeat(image, 3, axis=-1)
        elif ch == 2:
            image = jnp.concatenate([image, jnp.zeros_like(image[..., :1])], axis=-1)

        assert image.shape[-1] == 3

        # ensure input size is multiple of 32
        height = ((orig_height - 1) // 32 + 1) * 32
        width = ((orig_width - 1) // 32 + 1) * 32
        image = jnp.pad(
            image, [[0, height - orig_height], [0, width - orig_width], [0, 0]]
        )

        # backbone
        encoder_features, features = self._backbone(image, training=training)
        model_output = dict(
            encoder_features=encoder_features,
            decoder_features=features,
        )

        # detection
        scaled_gt_locations = (
            gt_locations / jnp.array([height, width])
            if gt_locations is not None
            else None
        )
        model_output.update(
            self._lpn(
                inputs=features,
                scaled_gt_locations=scaled_gt_locations,
                training=training,
            )
        )
        model_output.update(
            self._detector(
                scores=model_output["lpn_scores"],
                regressions=model_output["lpn_regressions"],
                gt_locations=gt_locations,
                training=training,
            )
        )

        # segmentation
        locations = model_output["training_locations" if training else "pred_locations"]
        scaled_locs = locations / jnp.array([height, width])
        model_output.update(
            self._segmentor(
                features=features,
                locations=scaled_locs,
            )
        )

        return model_output


class LacssWithHelper(nn.Module):
    cfg: dict = field(default_factory=dict)
    aux_edge_cfg: tp.Optional[dict] = None
    aux_fg_cfg: tp.Optional[dict] = None

    def setup(self):
        self._lacss = Lacss(**self.cfg)
        if self.aux_edge_cfg is not None:
            self._aux_edge_module = AuxInstanceEdge(**self.aux_edge_cfg)
        if self.aux_fg_cfg is not None:
            self._aux_fg_module = AuxForeground(**self.aux_fg_cfg)

    def __call__(self, image, gt_locations=None, category=None, *, training=None):

        preds = self._lacss(image, gt_locations, training=training)

        if self.aux_edge_cfg is not None:
            preds.update(
                dict(edge_pred=self._aux_edge_module(image, category=category))
            )
        if self.aux_fg_cfg is not None:
            preds.update(dict(fg_pred=self._aux_fg_module(image, category=category)))

        return preds



from typing import Dict, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np

from .boxes import box_iou_similarity

""" Various functions deals with segmentation pathces
    All functions here takes unbatched input. Use vmap to convert to batched data
"""


def gather_patches(features, locations, patch_size: int) -> tuple:
    """extract feature patches according to a list of locations
    Args:
        features: [H,W,C] standard 2D feature map
        locations: [N, 2] float32, scaled 0..1
        patch_size: int
    Returns:
        patches: [N, patch_size, patch_size, C]
        y0: [N]: y0 coordinates of patches
        x0: [N]: x0 coordinates of patches
    """
    height, width, _ = features.shape

    locations *= jnp.array([height, width])
    # i_locations = (locations + .5).astype(int)
    i_locations = locations.astype(int)
    i_locations_x = jnp.clip(i_locations[:, 1], 0, width - 1)
    i_locations_y = jnp.clip(i_locations[:, 0], 0, height - 1)
    yy, xx = jnp.mgrid[:patch_size, :patch_size] - patch_size // 2
    xx += i_locations_x[:, None, None]
    yy += i_locations_y[:, None, None]
    remainder = locations - jnp.stack([i_locations_y, i_locations_x], axis=-1)

    # padding to avoid out-of-bound
    padding_size = patch_size // 2 + 1
    paddings = [
        [padding_size, padding_size],
        [padding_size, padding_size],
        [0, 0],
    ]
    padded_features = jnp.pad(features, paddings)

    patches = padded_features[yy + padding_size, xx + padding_size, :]

    return patches, yy[:, 0, 0], xx[:, 0, 0], remainder


def _get_patch_data(pred):
    if isinstance(pred, dict):
        patches = pred["instance_output"]
        yc = pred["instance_yc"]
        xc = pred["instance_xc"]
    else:
        patches, yc, xc = pred

    if patches.ndim > yc.ndim:
        patches = patches.squeeze(-1)

    return patches, yc, xc


def bboxes_of_patches(
    pred: Union[Sequence, Dict], threshold: float = 0.5
) -> jnp.ndarray:
    """
    Args:
        pred: either a tuple or a dict containing three arrays:
            instance_outputs: [n, patch_size, patch_size, 1] float
            instance_y_coords: [n, patch_size, patch_size] y patch coords
            instance_x_coords: [n, patch_size, patch_size] x patch coords
        threshold: float
    Returns:
        bboxes: [n, 4] int
        bboox for empty patches are filled with -1
    """
    patches, yy, xx = _get_patch_data(pred)

    _, d0, d1 = patches.shape
    row_mask = jnp.any(patches > threshold, axis=1)
    col_mask = jnp.any(patches > threshold, axis=2)

    min_col = row_mask.argmax(axis=1)
    max_col = d1 - row_mask[:, ::-1].argmax(axis=1)

    min_row = col_mask.argmax(axis=1)
    max_row = d0 - col_mask[:, ::-1].argmax(axis=1)

    min_row += yy[:, 0, 0]
    max_row += yy[:, 0, 0]
    min_col += xx[:, 0, 0]
    max_col += xx[:, 0, 0]

    is_valid = row_mask.any(axis=1, keepdims=True)
    bboxes = jnp.stack([min_row, min_col, max_row, max_col], axis=-1)
    bboxes = jnp.where(is_valid, bboxes, -1)

    return bboxes


def indices_of_patches(
    pred: Union[Sequence, Dict],
    input_size: Tuple[int, int] = None,
    threshold: float = 0.5,
) -> tuple:
    """
    Args:
        pred: model output (unbatched):
        threshold: float
    Returns:
        data for MaskIndices
    """
    patches, yy, xx = _get_patch_data(pred)

    indices = patches >= threshold
    yy = yy[indices]
    xx = xx[indices]
    rowids = jnp.argwhere(indices)[:, 0]

    if input_size is not None:
        height, width = input_size
        valid_coords = ((yy >= 0) & (yy < height)) & ((xx >= 0) & (xx < width))
        yy = yy[valid_coords]
        xx = xx[valid_coords]
        rowids = rowids[valid_coords]

    return jnp.stack([yy, xx, rowids], axis=-1)


def iou_patches_and_labels(pred, labels, BLOCK_SIZE=128):
    """Compute iou between two sets of segmentations.
    The first set is defined by patches; the second by image labels
    Args:
        pred: model output (unbatched):
        threshold: float
    Returns:
        [n, m] iou values.
    """
    patches, yc, xc = _get_patch_data(pred)
    patches = patches >= 0.5
    pred_areas = jnp.count_nonzero(patches, axis=(-1, -2))

    # this rely on JAX's out-of-bound indexing behavior
    padded_labels = jnp.pad(labels, [[1, 1], [1, 1]], constant_values=-1)
    gt_patches = padded_labels[yc + 1, xc + 1]
    max_indices = ((labels.max() - 1) // BLOCK_SIZE + 1) * BLOCK_SIZE
    all_gt_areas = jnp.count_nonzero(
        labels[:, :, None] == jnp.arange(max_indices) + 1, axis=(0, 1)
    )

    ious = []
    # FIXME change to scan
    for k in range(1, labels.max() + 1, BLOCK_SIZE):
        gt_p = gt_patches == jnp.arange(k, k + BLOCK_SIZE).reshape(
            -1, 1, 1, 1
        )  # [B, N, s, s]
        gt_areas = all_gt_areas[k - 1 : k + BLOCK_SIZE - 1]
        intersect = jnp.count_nonzero(gt_p & patches, axis=(-1, -2))  # [B, N]
        ious.append(intersect / (pred_areas + gt_areas[:, None] - intersect + 1.0e-8))

    ious = jnp.concatenate(ious, axis=0).transpose()  # [N, B * b]
    ious = ious[:, : labels.max()]
    return ious


def patches_to_segmentations(pred, input_size, threshold=0.5):
    """expand patches to the full image size
    Args:
        pred: model output (unbatched):
        input_size: tuple(int, int)
        threshold: float
    Returns:
        segmentations: [n, height, width]
    """
    patches, yc, xc = _get_patch_data(pred)
    n_patches, patch_size, _ = yc.shape

    page_nums = jnp.arange(n_patches)
    segms = jnp.zeros((n_patches,) + input_size)
    segms = segms.at[page_nums[:, None, None], yc, xc].set(patches)

    return (segms >= threshold).astype(int)


def patches_to_label(
    pred, input_size, mask=None, score_threshold=0.5, threshold=0.5, min_cell_area=0.0
):
    """convert patch output to the image label
    Args:
        pred: model output (unbatched):
        input_size: a int tuple of [heght, width]
        mask: boolean indicators of which cell to display, default is None (all cells)
        score_threshold: otional, min_score to be plotted, default .5
        threshold: optional, output threshold,  default .5
        min_cell_area: optional min cell area to be plotted, default 0.
    Returns:
        label: [height, width]
    """
    label = jnp.zeros(input_size)
    pr = pred["instance_output"] > threshold
    n_patches, patch_size, _ = pr.shape

    if mask is None:
        mask = jnp.ones([n_patches], dtype=bool)
    mask &= pred["instance_mask"].squeeze(axis=(1, 2))
    if score_threshold > 0 and not "training_locations" in pred:
        mask &= pred["pred_scores"] >= score_threshold
    mask &= jnp.count_nonzero(pr, axis=(1, 2)) > min_cell_area

    pr = (pr > threshold).astype(int) * jnp.arange(1, pr.shape[0] + 1)[:, None, None]
    pr = jnp.where(mask[:, None, None], pr, 0)
    pr = jnp.where(pr == 0, 0, pr.max() + 1 - pr)

    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    label = label.at[yc, xc].max(pr)

    return label


def ious_of_patches_from_same_image(pred, threshold=0.5):
    """
    Args:
        pred: model output without batch dim
        threshold: output threshold, default to 0.5
    Returns:
        ious: upper triagle matrix. mask ious of cells indicated by yc and xc
    """
    bboxes = bboxes_of_patches(pred)
    box_ious = box_iou_similarity(bboxes, bboxes)
    box_ious = np.triu(box_ious, 1)
    # n_detections = jnp.count_nonzero(pred['instance_mask'])
    # box_ious = box_ious[:n_detections, :n_detections]

    cy, cx = np.where(box_ious > 0)
    pad_size = pred["instance_output"].shape[-1]

    patches = np.asarray(pred["instance_output"]) >= threshold
    patches_y = patches[cy]
    patches_x = patches[cx]
    plt_to = np.zeros([len(cy), pad_size * 3, pad_size * 3], dtype=bool)
    dy = (
        np.asarray(pred["instance_yc"])[cy, 0, 0]
        - np.asarray(pred["instance_yc"])[cx, 0, 0]
        + pad_size
    )
    dx = (
        np.asarray(pred["instance_xc"])[cy, 0, 0]
        - np.asarray(pred["instance_xc"])[cx, 0, 0]
        + pad_size
    )

    for k in range(len(cy)):
        plt_to[k, dy[k] : dy[k] + pad_size, dx[k] : dx[k] + pad_size] = patches_y[k]
    plt_to = plt_to[:, pad_size : pad_size * 2, pad_size : pad_size * 2]
    unions = np.count_nonzero(plt_to & patches_x, axis=(1, 2))

    areas_y = np.count_nonzero(patches_y, axis=(1, 2))
    areas_x = np.count_nonzero(patches_x, axis=(1, 2))

    ious = unions / (areas_x + areas_y - unions + 1e-8)

    mask_ious = box_ious
    mask_ious[(cy, cx)] = ious

    return mask_ious


def non_max_suppress_predictions(pred, iou_threshold=0.6):
    """
    Args:
        pred: model output without batch dim
        ious_threshold: default 0.6
    Returns:
        mask: boolean mask of cells not supressed
    """
    ious = ious_of_patches_from_same_image(pred)
    mask = ious > iou_threshold

    cnt = 0
    while np.count_nonzero(mask) != cnt:
        cnt = jnp.count_nonzero(mask)
        can_suppress_others = ~mask.any(axis=0)
        suppressed = (mask & can_suppress_others[:, None]).any(axis=0)
        mask = mask & ~suppressed[:, None]

    return ~mask.any(axis=0)



from .array_adapter import ArrayDataAdapter
from .data_handler import DataHandler
from .dataset import DataLoader, DataLoaderAdapter, Dataset
from .generator_adapter import GeneratorDataAdapter
from .list_adapter import ListsOfScalarsDataAdapter
from .utils import (
    map_append,
    map_structure,
    train_validation_split,
    unpack_x_y_sample_weight,
)

try:
    from .tf_dataset_adapter import TFDatasetAdapter
except ImportError:
    TFDatasetAdapter = None
try:
    from .torch_dataloader_adapter import TorchDataLoaderAdapter
except ImportError:
    TorchDataLoaderAdapter = None

ALL_ADAPTER_CLS = [
    ArrayDataAdapter,
    GeneratorDataAdapter,
    ListsOfScalarsDataAdapter,
    DataLoaderAdapter,
]



from functools import partial
from typing import Dict, List, Sequence, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..ops import gather_patches
from .common import SpatialAttention


class _Encoder(nn.Module):
    n_ch: int = 16
    patch_size: int = 48
    resolution: int = 8
    encoding_ch: int = 4

    @nn.compact
    def __call__(self, feature, loc):

        patch_center, _, _, _ = gather_patches(feature, loc, patch_size=2)
        encodings = patch_center.mean(axis=(-2, -3))

        dim = self.encoding_ch * self.resolution * self.resolution
        encodings = nn.Dense(dim)(encodings)
        encodings = jax.nn.relu(encodings)
        encodings = nn.Dense(dim)(encodings)
        encodings = jax.nn.relu(encodings)

        encoding_shape = (-1, self.resolution, self.resolution, self.encoding_ch)
        encodings = nn.Dense(dim)(encodings).reshape(encoding_shape)
        encodings = jax.image.resize(
            encodings,
            [encodings.shape[0], self.patch_size, self.patch_size, self.encoding_ch],
            "linear",
        )

        encodings = nn.Conv(self.n_ch, (1, 1), use_bias=False)(encodings)

        return encodings


class Segmentor(nn.Module):
    """
    Args:
    conv_spec: conv_block definition, e.g. ((64, 64, 64), (16, 16))
    instance_crop_size: crop size, int
    feature_scale_ratio: 1,2 or 4
    with_attention: T/F whether use spatial attention layer
    learned_encoding: Use hard-coded position encoding or not
    """

    feature_level: int = 1
    conv_spec: Tuple[Sequence[int], Sequence[int]] = (
        (64, 64, 64),
        (16,),
    )
    instance_crop_size: int = 96
    use_attention: bool = False
    learned_encoding: bool = False
    n_cls: int = -1

    # if not feature_level in (0,1,2):
    #     raise ValueError('feature_level should be 1,2 or 0')

    @nn.compact
    def __call__(self, features: dict, locations: jnp.ndarray) -> tuple:
        """
        Args:
            features: {'lvl' [H, W, C]} feature dictionary
            locations: [N, 2]  scaled 0..1
        outputs:
            instance_output: [N, crop_size, crop_size, 1]
            yc: [N, crop_size, crop_size]
            xc: [N, crop_size, crop_size]
        """
        crop_size = self.instance_crop_size
        lvl = self.feature_level
        scale = 2**lvl
        conv_spec = self.conv_spec
        patch_size = crop_size // scale

        x = features[str(lvl)]
        # h, w, _ = x.shape

        # feature convs
        for n_ch in conv_spec[0]:
            x = nn.Conv(n_ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=n_ch)(x[None, ...])[0]
            x = jax.nn.relu(x)

        # mixing features with pos_encodings
        n_ch = conv_spec[1][0]
        x = nn.Conv(n_ch, (3, 3), use_bias=False)(x)
        patches, y0, x0, _ = gather_patches(x, locations, patch_size=patch_size)

        if not self.learned_encoding:
            encodings = (
                jnp.mgrid[:patch_size, :patch_size] / patch_size - 0.5
            ).transpose(1, 2, 0)
            encodings = nn.Conv(n_ch, (1, 1), use_bias=False)(encodings)
        else:
            encodings = _Encoder(n_ch, patch_size)(features[str(lvl)], locations)
        patches += encodings
        mask = jnp.expand_dims((locations >= 0).any(axis=-1), (1, 2))
        patches = jax.nn.relu(patches)

        if self.use_attention:
            patches = SpatialAttention()(patches)

        # patche convs
        for n_ch in conv_spec[1][1:]:
            patches = nn.Conv(n_ch, (3, 3))(patches)
            patches = jax.nn.relu(patches)

        # outputs
        if scale == 1:
            logits = nn.Conv(1, (3, 3))(patches)
        else:
            logits = nn.ConvTranspose(1, (2, 2), strides=(2, 2))(patches)
            if scale == 4:
                logits = jax.image.resize(
                    logits, patches.shape[:1] + (crop_size, crop_size, 1), "linear"
                )

        logits = logits.squeeze(-1)
        outputs = jax.nn.sigmoid(logits)

        # indicies
        yc, xc = jnp.mgrid[:crop_size, :crop_size]
        yc = yc + y0[:, None, None] * scale
        xc = xc + x0[:, None, None] * scale

        # clear invalid locations
        outputs = jnp.where(mask, outputs, 0)
        logits = jnp.where(mask, logits, 0)
        xc = jnp.where(mask, xc, -1)
        yc = jnp.where(mask, yc, -1)

        return dict(
            instance_output=outputs,
            instance_yc=yc,
            instance_xc=xc,
            instance_logit=logits,
            instance_mask=mask,
        )



import numpy as np
import tensorflow as tf


def crop_or_pad_to_target_size(image, target_height, target_width):
    height = tf.shape(image)[0]
    width = tf.shape(image)[1]
    # n_ch = tf.shape(image)[2]

    d_h = height - target_height
    d_w = width - target_width
    if d_h > 0:
        h0 = tf.random.uniform((), 0, d_h + 1, tf.int32)
        image = tf.image.crop_to_bounding_box(image, h0, 0, target_height, width)
    else:
        h0 = 0
        image = tf.image.pad_to_bounding_box(image, 0, 0, target_height, width)
    if d_w > 0:
        w0 = tf.random.uniform((), 0, d_w + 1, tf.int32)
        image = tf.image.crop_to_bounding_box(image, 0, w0, target_height, target_width)
    else:
        w0 = 0
        image = tf.image.pad_to_bounding_box(image, 0, 0, target_height, target_width)

    image = tf.ensure_shape(image, [target_height, target_width, None])
    return image, h0, w0


def parse_train_data_func_full_annotation(
    data, augment=True, min_mask_area=16, target_height=512, target_width=512
):
    image = data["image"]
    masks = data["mask_indices"]
    mask_row_ids = masks.value_rowids()
    mask_values = masks.values

    target_width = int(target_width)
    target_height = int(target_height)

    image, h0, w0 = crop_or_pad_to_target_size(image, target_height, target_width)
    mask_values = mask_values - tf.cast([h0, w0], mask_values.dtype)
    if augment:
        if tf.random.uniform(()) >= 0.5:
            image = tf.image.flip_left_right(image)
            mask_values = mask_values * [1, -1] + [0, target_width - 1]
        if tf.random.uniform(()) >= 0.5:
            image = tf.image.flip_up_down(image)
            mask_values = mask_values * [-1, 1] + [target_height - 1, 0]

    valid_rows = tf.logical_and(
        tf.logical_and(mask_values[:, 0] >= 0, mask_values[:, 0] < target_height),
        tf.logical_and(mask_values[:, 1] >= 0, mask_values[:, 1] < target_width),
    )
    mask_values = tf.boolean_mask(mask_values, valid_rows)
    mask_row_ids = tf.boolean_mask(mask_row_ids, valid_rows)
    masks = tf.RaggedTensor.from_value_rowids(mask_values, mask_row_ids)
    mask_row_lengths = masks.row_lengths()
    valid_masks = mask_row_lengths >= min_mask_area
    masks = tf.ragged.boolean_mask(masks, valid_masks)

    locations = tf.cast(tf.reduce_mean(masks, axis=1), tf.float32)

    row_ids = masks.value_rowids() + 1
    labels = tf.zeros([target_height, target_width], row_ids.dtype)
    labels = tf.tensor_scatter_nd_update(labels, masks.values, row_ids)

    return {
        "image": image,
        "locations": locations,
        "mask_labels": labels,
    }


def parse_train_data_func(
    data, augment=True, size_jitter=None, target_height=512, target_width=512
):
    image = data["image"]
    binary_mask = tf.expand_dims(data["binary_mask"], -1)
    locations = data["locations"]

    img_and_label = tf.concat([image, tf.cast(binary_mask, tf.float32)], -1)
    if size_jitter is not None:
        h = tf.cast(tf.shape(img_and_label)[-3], tf.float32)
        w = tf.cast(tf.shape(img_and_label)[-2], tf.float32)
        scaling = tf.random.uniform([], size_jitter[0], size_jitter[1])
        img_and_label = tf.image.resize(
            img_and_label, [int(h * scaling), int(w * scaling)], antialias=True
        )
        locations = locations * scaling

    img_and_label, h0, w0 = crop_or_pad_to_target_size(
        img_and_label, target_height, target_width
    )
    locations = locations - tf.cast([h0, w0], tf.float32)

    if augment:
        if tf.random.uniform([]) >= 0.5:
            img_and_label = tf.image.flip_left_right(img_and_label)
            locations = locations * [1.0, -1.0] + [0, target_width]
        if tf.random.uniform([]) >= 0.5:
            img_and_label = tf.image.flip_up_down(img_and_label)
            locations = locations * [-1.0, 1.0] + [target_height, 0]

        if target_height == target_width:
            if tf.random.uniform([]) >= 0.5:
                img_and_label = tf.image.transpose(img_and_label)
                locations = locations[:, ::-1]

    image = img_and_label[..., :-1]
    if size_jitter is not None:
        binary_mask = tf.cast(img_and_label[..., -1] > 0.5, tf.float32)
    else:
        binary_mask = img_and_label[..., -1]

    # remove out-of-bound locations
    is_valid = tf.logical_and(
        tf.logical_and(locations[:, 0] > 1, locations[:, 0] < target_height - 1),
        tf.logical_and(locations[:, 1] > 1, locations[:, 1] < target_width - 1),
    )
    locations = tf.boolean_mask(locations, is_valid)

    # if augment:
    #     image = tf.image.random_brightness(image, 0.3)
    #     image = tf.image.random_contrast(image, 0.7, 1.3)

    return {
        "image": image,
        "locations": locations,
        "binary_mask": binary_mask,
    }


def parse_test_data_func(data, dim_multiple=32):
    image = data["image"]
    binary_mask = tf.expand_dims(data["binary_mask"], -1)
    bboxes = data["bboxes"]
    locations = data["locations"]
    masks = data["mask_indices"]

    height = tf.shape(image)[0]
    width = tf.shape(image)[1]
    target_height = (height - 1) // dim_multiple * dim_multiple + dim_multiple
    target_width = (width - 1) // dim_multiple * dim_multiple + dim_multiple

    image = tf.image.pad_to_bounding_box(image, 0, 0, target_height, target_width)
    binary_mask = tf.cast(
        tf.image.pad_to_bounding_box(binary_mask, 0, 0, target_height, target_width),
        tf.float32,
    )
    binary_mask = tf.squeeze(binary_mask, -1)

    scaling = data["scaling"] if "scaling" in data else tf.constant(1.0, tf.float32)

    row_ids = masks.value_rowids() + 1
    labels = tf.zeros([target_height, target_width], row_ids.dtype)
    labels = tf.tensor_scatter_nd_update(labels, masks.values, row_ids)

    return {
        "image": image,
        "bboxes": bboxes,
        "locations": locations,
        "img_id": data["img_id"],
        "scaling": scaling,
        "binary_mask": binary_mask,
        "mask_indices": masks,
        "mask_labels": labels,
    }



from functools import partial

import jax
from jax.scipy.signal import convolve

jnp = jax.numpy


def sorbel_edges(image):
    """Returns a tensor holding Sobel edge maps.
    >>> image = random.uniform(key, shape=[3, 28, 28])
    >>> sobel = sobel_edges(image)
    >>> sobel_y = sobel[0, :, :, :] # sobel in y-direction
    >>> sobel_x = sobel[1, :, :, :] # sobel in x-direction
    ```
    Args:
        image: [n, h, w]
    Returns:
        Tensor holding edge maps for each channel. [2, n, h, w]
    """

    kernels = jnp.array(
        [[[-1, -2, -1], [0, 0, 0], [1, 2, 1]], [[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]]]
    )

    sorbel_filter = jax.vmap(
        jax.vmap(
            partial(convolve, mode="valid"),
            in_axes=[0, None],
        ),
        in_axes=[None, 0],
    )

    image = jnp.pad(image, [[0, 0], [1, 1], [1, 1]], mode="reflect")
    output = sorbel_filter(image, kernels)

    return output



import pathlib
import typing as tp
from functools import lru_cache, partial

import cloudpickle
import flax.linen as nn
import jax
import jax.numpy as jnp
from flax.training.train_state import TrainState
from optax import GradientTransformation

from ..utils import Inputs, _get_name
from . import strategy
from .data import *
from .loss import Loss, LossLog

# so that multiple calls return the same obj
# this avoids JIT when supplying partial func as args
_cached_partial = lru_cache(partial)


class Trainer:
    def __init__(
        self,
        model: nn.Module,
        losses: tp.Optional[tp.Union[tp.Sequence[Loss], Loss]] = None,
        optimizer: GradientTransformation = None,
        seed: int = 42,
        strategy: type = strategy.JIT,
    ):
        self.model = model
        self.losses = (losses,) if isinstance(losses, Loss) else losses
        self.seed = seed if isinstance(seed, jnp.ndarray) else jax.random.PRNGKey(seed)
        self.optimizer = optimizer

        self._strategy = strategy
        self._initialized = False

        self.reset()

    def reset(self):
        self.loss_logs = tuple(LossLog(loss) for loss in self.losses)

    @property
    def initialized(self):
        return self._initialized

    def initialize(self, dataset, tx: GradientTransformation = None):

        if tx is None:
            tx = self.optimizer

        if not self._initialized:
            peek = next(dataset())
            inputs, _, _ = unpack_x_y_sample_weight(peek)

            self.seed, key = jax.random.split(self.seed)
            self.state = self._strategy.init_fn(key, self.model, inputs, tx)

        self.reset()
        self._initialized = True

    def __call__(self, inputs, *, strategy=None, **kwargs):
        if strategy is None:
            strategy = self._strategy

        predict_fn = strategy.predict

        state = self.state.replace(
            apply_fn=_cached_partial(self.state.apply_fn, **kwargs)
        )
        preds = predict_fn(state, inputs)

        return preds

    def compute_loss_log(self):
        return {
            loss_log.loss_fn.name: loss_log.compute() for loss_log in self.loss_logs
        }

    def train(self, dataset, strategy=None, rng_cols=None, **kwargs):
        if strategy is None:
            strategy = self._strategy

        if not self._initialized:
            raise ValueError("Try to run uninitialized trainer")

        self.reset()

        self.seed, seed = jax.random.split(self.seed)
        for step, data in enumerate(dataset()):
            inputs, labels, _ = unpack_x_y_sample_weight(data)

            if rng_cols is not None:
                key = jax.random.fold_in(seed, step)
                keys = jax.random.split(key, len(rng_cols))
                rngs = {name: k for name, k in zip(rng_cols, keys)}
            else:
                rngs = None

            train_fn = strategy.train_step

            state = self.state.replace(
                apply_fn=_cached_partial(self.state.apply_fn, **kwargs)
            )
            state, self.loss_logs, preds = train_fn(
                state, self.loss_logs, inputs, labels, rngs
            )
            self.state = state.replace(apply_fn=self.state.apply_fn)

            batch_logs = self.compute_loss_log()

            yield batch_logs

    def checkpoint(self, path):
        if isinstance(path, str):
            path = pathlib.Path(path)

        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "wb") as f:
            cloudpickle.dump(self, f)

    @staticmethod
    def from_checkpoint(path):
        if isinstance(path, str):
            path = pathlib.Path(path)

        try:
            _bytes = path.read_bytes()
        except BaseException as e:
            raise OSError(f"Could not load the checkpoint. Got exception: {e}")

        return cloudpickle.loads(_bytes)

    def test(self, dataset, metrics, strategy=None):
        if strategy is None:
            strategy = self._strategy

        try:
            iter(metrics)
        except TypeError:
            metrics = [metrics]

        # self.initialize(dataset, strategy)

        for data in dataset():
            inputs, labels, _ = unpack_x_y_sample_weight(data)
            predict_fn = strategy.predict
            preds = predict_fn(self.state, inputs)
            kwargs = dict(
                inputs=inputs,
                preds=preds,
                **labels,
            )
            for m in metrics:
                m.update(**kwargs)

            yield metrics

    def test_and_compute(self, *args, **kwargs):
        for metrics in self.test(*args, **kwargs):
            pass
        return {_get_name(m): m.compute() for m in metrics}

    @property
    def params(self):
        return self.state.params



import typing as tp
from abc import abstractmethod


class Metric:
    @abstractmethod
    def update(self, *args, **kwargs):
        ...

    @abstractmethod
    def compute(self, *args, **kwargs) -> tp.Dict:
        ...



import jax
import jax.numpy as jnp
import numpy as np

from ..ops import *
from ..train.metric import Metric

"""
metrics classes for computing coco-style AP metrics.

BE CAREFUL making changes here. Very easy to make a mistake and resulting mismatch with
coco algorithm. Be sure to validate against coco-evaluation before commit changes.
"""


def _unique_location_matching(similarity_matrix, threshold):
    """Perform matching based on similarity_matrix.
    This is different from the matchre functions in ops/*.py in that the matching
    will be unique: each column will match at most one row
    Args:
      similarity_matrix: [N, K] tensor, rows are presorted based on scores
      threshold: minimal value to be considered as a match
    Returns:
      matches: [N,] indices[0..K)  of the match for each row
      indicators: [N,] indicator (1/0) value for each match
    """

    matches = []
    indicators = []
    similarity_matrix = np.copy(similarity_matrix)
    n, _ = similarity_matrix.shape
    for i_row in range(n):
        row = similarity_matrix[i_row]
        row_max_indice = row.argmax()
        row_max = row[row_max_indice]
        matches.append(row_max_indice)
        if row_max > threshold:
            indicators.append(1)
            similarity_matrix[:, row_max_indice] = -1
        else:
            indicators.append(0)

    return np.array(matches, np.int32), np.array(indicators, np.int32)


def np_compute_ap(similarity_matrix, thresholds):
    # avoid edge cases
    _, k = similarity_matrix.shape
    if k == 0:
        return np.zeros([len(thresholds)], np.float32)

    apmks = []
    for th in thresholds:
        _, indicators = _unique_location_matching(similarity_matrix, th)
        p_k = np.cumsum(indicators) / (np.arange(len(indicators)) + 1)
        apmks.append(np.sum(p_k * indicators) / k)

    return np.array(apmks, np.float32)


class AP(Metric):
    """compute AP based on similarity_matrix and score.
    These are numpy functions
    Usage:
      m = MeanAP([threshold_1, threshold_2,...])
      m.update_states(similarity_matrix, scores)
      ...
      m.result()
    """

    def __init__(self, thresholds=[0.5], coco_style=False):
        self.thresholds = thresholds
        self.coco_style = coco_style
        self.name = "AP"
        self.reset()

    def update(self, sm, scores):
        self.cell_counts += sm.shape[1]
        self.scores.append(scores)
        for th, indicators in zip(self.thresholds, self.indicator_list):
            _, ind = _unique_location_matching(sm, th)
            indicators.append(ind)

        self._result = None

    def compute(self):
        if self._result is not None:
            return self._result

        scores = np.concatenate(self.scores)
        indices = np.argsort(scores)
        aps = []
        for indicators in self.indicator_list:
            indicators = np.concatenate(indicators)
            indicators = indicators[indices[::-1]]
            p_k = np.cumsum(indicators) / (np.arange(len(indicators)) + 1)
            if self.coco_style:
                p_k = np.maximum.accumulate(p_k[::-1])[::-1]
            aps.append(np.sum(p_k[indicators == 1]) / self.cell_counts)

        self._result = np.array(aps, dtype=float)

        return self._result

    def reset(self):
        self.scores = []
        self.cell_counts = 0
        self.indicator_list = [[] for _ in range(len(self.thresholds))]
        self._result = np.array([-1.0] * len(self.thresholds))


class LoiAP(AP):
    def __init__(self, thresholds=[0.5], **kwargs):
        super().__init__([th * th for th in thresholds], **kwargs)
        self.name = "LoiAP"

    def update(self, preds, gt_locations, **kwargs):
        scores = preds["pred_scores"]
        pred_locations = preds["pred_locations"]
        n_batch = gt_locations.shape[0]
        for k in range(n_batch):
            score = scores[k]
            pred = pred_locations[k]
            gt = gt_locations[k]
            row_mask = score > 0
            col_mask = (gt >= 0).all(axis=-1)
            dist2 = ((pred[:, None, :] - gt[:, :]) ** 2).sum(axis=-1)

            dist2 = np.asarray(1.0 / dist2)
            dist2 = dist2[row_mask][:, col_mask]
            score = np.asarray(score)[row_mask]

            super().update(dist2, score)


class BoxAP(AP):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "BoxAP"

    def update(self, preds, gt_boxes, **kwargs):
        pred_boxes = jax.vmap(bboxes_of_patches)(preds)
        box_ious = box_iou_similarity(pred_boxes, gt_boxes)
        scores = preds["pred_scores"]

        for score, gt_box, iou in zip(scores, gt_boxes, box_ious):
            iou = iou[score >= 0][:, (gt_box >= 0).all(axis=-1)]
            super().update(iou, score[score >= 0])



from typing import Dict, List, Sequence, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp

from .common import *


class AuxInstanceEdge(nn.Module):
    conv_spec: Sequence[int] = (24, 64, 64)
    n_groups: int = 1
    share_weights: bool = False

    @nn.compact
    def __call__(
        self, x: jnp.ndarray, *, category: Optional[int] = None
    ) -> jnp.ndarray:
        for n in self.conv_spec:
            x = nn.Conv(n, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=None, group_size=1, use_scale=False)(
                x[None, ...]
            )[0]
            x = jax.nn.relu(x)

        x = nn.Conv(self.n_groups, (3, 3))(x)
        x = x[..., category if category is not None else 0]

        # x = jax.nn.sigmoid(x)

        return x


class AuxForeground(nn.Module):
    conv_spec: Sequence[int] = (24, 64, 64)
    n_groups: int = 1
    share_weights: bool = False

    @nn.compact
    def __call__(
        self, x: jnp.ndarray, *, category: Optional[int] = None
    ) -> jnp.ndarray:
        orig = x
        n_ch = self.conv_spec[0]
        x = nn.Conv(n_ch, (3, 3), strides=(2, 2), use_bias=False)(x)
        x = nn.GroupNorm(num_groups=n_ch, use_scale=False)(x[None, ...])[0]
        x = jax.nn.relu(x)
        for n_ch in self.conv_spec[1:]:
            x = nn.Conv(n_ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=n_ch, use_scale=False)(x[None, ...])[0]
            x = jax.nn.relu(x)

        x = SpatialAttention()(x)
        x = jax.image.resize(x, orig.shape[:-1] + x.shape[-1:], "cubic")
        x = jnp.concatenate([orig, x], axis=-1)

        x = nn.Conv(self.n_groups, (3, 3))(x)
        x = x[..., category if category is not None else 0]

        # x = jax.nn.sigmoid(x)

        return x



import dataclasses
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np


def sample_unselected(selected, p, p_miss, *, key):
    """
    Args:
        selected: [n_sample, n_target]
        p: [n_target]
        p_miss: [1]
        n: int 1 or 2
        key: rng
    Returns:
        selection: [n_sample, n]
    """

    @jax.vmap
    def _mapped_choice(keys, weights):
        n_target = weights.size
        return jax.random.choice(
            keys,
            n_target,
            shape=[
                1,
            ],
            replace=False,
            p=weights,
        )

    n_sample, n_target = selected.shape

    total_mass = p_miss + p.sum()
    p_miss = p_miss / total_mass
    p = p / total_mass
    w = p * (1 - selected)  # [n_samples, n_targets]
    w = jnp.pad(w, [[0, 0], [0, 1]], constant_values=p_miss)
    w = w / w.sum(axis=-1, keepdims=True)  # noramlze each row
    assert w.shape == (n_sample, n_target + 1)

    keys = jax.random.split(key, n_sample)
    selection = _mapped_choice(keys, w)

    # formatting issue
    selection = jnp.where(selection == n_target, -1, selection)

    return selection


def update_history(history, selected, selection, k):
    n_sample = history.shape[0]
    history = history.at[:, k].set(selection[:, 0])
    # history = jnp.concatenate([history, selection], axis=-1)
    selected = jnp.pad(
        selected, [[0, 0], [0, 1]]
    )  # pad an extra column to catch both -1
    selected = selected.at[jnp.arange(n_sample), selection[:, 0]].set(1)
    selected = selected[:, :-1]  # remove padding

    return history, selected


def sample_and_update_history(history, selected, k, p, p_miss, *, key):
    selection = sample_unselected(selected, p, p_miss, key=key)
    history, selected = update_history(history, selected, selection, k)
    return history, selected


def get_sample_weight(selected, p, p_miss):
    total_mass = p_miss + p.sum()
    p_miss = p_miss / total_mass
    p = p / total_mass

    w = (p * (1 - selected)).sum(axis=-1) + p_miss
    w = w / w.sum()
    return w


def resample_history(history, selected, p, p_miss, *, key):
    n_sample = selected.shape[0]
    w = get_sample_weight(selected, p, p_miss)
    # print(w.shape)
    c = jax.random.choice(key, n_sample, p=w, shape=[n_sample])
    history = history[c]
    selected = selected[c]

    return history, selected


def seq_choice(key, wt, w_miss, w_nd, n_sample, resample=False):
    """
    Args
        key: rng
        wt: ndarray [n_source, n_target]
        w_miss: [n_source]
        w_nd: [n_source]
        n_sample: int
        resample: bool, default False
    returns: a single sample
        history: [n_source]
        selected: [n_target]
    """

    n_source, n_target = wt.shape

    selected = jnp.zeros([n_sample, n_target], dtype=int)

    log_weight = jnp.zeros([n_sample])

    history = jnp.zeros(shape=[n_sample, n_source], dtype=int) - 1
    history_div = jnp.zeros(shape=[n_sample, n_source], dtype=int) - 1

    def _for_inner(k, state):
        history, selected, key, log_weight = state
        w = get_sample_weight(selected, wt[k], w_miss[k])
        if resample:
            key, ckey = jax.random.split(key)
            c = jax.random.choice(ckey, n_sample, p=w, shape=[n_sample])
            history = history[c]
            selected = selected[c]
        else:
            log_weight += jnp.log(w)

        key, ckey = jax.random.split(key)
        history, selected = sample_and_update_history(
            history, selected, k, wt[k], w_miss[k], key=ckey
        )

        return history, selected, key, log_weight

    history, selected, key, log_weight = jax.lax.fori_loop(
        0,
        n_source,
        lambda k, state: jax.lax.cond(
            jnp.any(wt[k] > 0),
            _for_inner,
            lambda _, state: state,
            k,
            state,
        ),
        (history, selected, key, log_weight),
    )

    if not resample:
        weight = jnp.exp(log_weight - log_weight.max())
        weight = weight / weight.sum()
        rs = jax.random.choice(key, n_sample, shape=[n_sample], p=weight)
        selected = selected[rs]
        history = history[rs]
        log_weight = jnp.zeros([n_sample])

    key, ckey = jax.random.split(key)
    div = jax.random.uniform(ckey, w_nd.shape) < 1 / (1 + w_nd)
    wt = jnp.where(div[:, None], wt, 0.0)
    # w_miss += w_nd * (wt.sum(axis=1) + w_miss)
    history_div, selected, key, log_weight = jax.lax.fori_loop(
        0,
        n_source,
        lambda k, state: jax.lax.cond(
            jnp.any(wt[k] > 0),
            _for_inner,
            lambda _, state: state,
            k,
            state,
        ),
        (history_div, selected, key, log_weight),
    )

    if resample:
        rs = 0
    else:
        weight = jnp.exp(log_weight - log_weight.max())
        weight = weight / weight.sum()
        rs = jax.random.choice(key, n_sample, p=weight)

    return history[rs], history_div[rs], selected[rs]


@partial(jax.jit, static_argnums=4)
def _sample_step(key, padded_wts, w_miss, padded_nd, n_sub_sample):
    print(padded_wts.shape)
    n_sample, _, n_target = padded_wts.shape
    k0, k1, k2, k3 = jax.random.split(key, 4)
    k1 = jax.random.split(k1, n_sample)
    k2 = jax.random.split(k2, n_sample)
    k3 = jax.random.split(k3, n_sample)

    # n_target = padded_wts.shape[-1]
    # p_div = 1 / (1 + padded_nd)
    # div = (jax.random.uniform(k0, [n_sample, n_sub_sample, n_target]) <= p_div[:, None, :]).astype(int)
    history, history_div, selected = jax.vmap(
        partial(seq_choice, n_sample=n_sub_sample)
    )(
        k1,
        padded_wts,
        w_miss,
        padded_nd,
    )

    reversed = (history == -1) & (history_div != -1)
    history = jnp.where(reversed, history_div, history)
    history_div = jnp.where(reversed, -1, history_div)

    return history, history_div, selected


def sample_step(key, selected, wts, w_miss, w_nd, n_sub_sample=256):
    """
    Args:
        key: a single key
        selected: [n_sample, n_source]
        wts: [n_source, n_target], pad with 0 will ensure all invalid rows return -1
        w_miss: [n_source]
        w_nd: [n_samples, n_source]
    Returns:
        history: [n_samples, n_source], -1 means missing
        history_div: [n_samples, n_source], -1 means no division
        selected: [n_samples, n_target], 1/0
    """
    ROWPADSIZE = 32
    COLPADSIZE = 32

    n_sample, _ = selected.shape
    n_source, n_target = wts.shape

    wts = np.array(wts)
    w_nd = np.array(w_nd)
    selected = np.asarray(selected).astype(bool)

    actual_n_source = selected.sum(-1)
    padto_source = (actual_n_source.max() - 1) // ROWPADSIZE * ROWPADSIZE + ROWPADSIZE
    padto_target = n_target // COLPADSIZE * COLPADSIZE + COLPADSIZE - 1

    padded_wts = np.stack(
        [
            np.pad(
                wts[selected[k]],
                [[0, padto_source - actual_n_source[k]], [0, padto_target - n_target]],
            )
            for k in range(n_sample)
        ]
    )
    padded_nd = np.stack(
        [
            np.pad(w_nd[k][selected[k]], [0, padto_source - actual_n_source[k]])
            for k in range(n_sample)
        ]
    )
    w_miss = np.stack(
        [
            np.pad(w_miss[selected[k]], [0, padto_source - actual_n_source[k]])
            for k in range(n_sample)
        ]
    )

    padded_history, padded_history_div, padded_selected = _sample_step(
        key, padded_wts, w_miss, padded_nd, n_sub_sample
    )
    padded_history = np.stack([padded_history, padded_history_div], axis=-1)
    history = np.zeros([n_sample, n_source, 2], dtype=int)
    for k in range(n_sample):
        history[k, selected[k], :] = padded_history[k, : actual_n_source[k], :]
    selected = padded_selected[:, :n_target]

    return history, selected


def post_process_step(history):
    selected0 = history["samples"][-2]["selected"]
    selected1 = history["samples"][-1]["selected"]
    link, div = np.moveaxis(history["samples"][-2]["next"], -1, 0)
    lifetime0 = history["samples"][-2]["lifetime"]
    yx1 = history["detections"][-1]["yx_next"]
    yx0 = history["detections"][-1]["yx"]
    id1 = history["detections"][-1]["id_next"]
    id0 = history["detections"][-1]["id"]

    n_target = yx1.shape[0]

    def _update_lifetime(lifetime0, link, div):
        lifetime = jnp.zeros([n_target + 1])
        link_no_div = jnp.where(div != -1, -1, link)
        lifetime = lifetime.at[link_no_div].set(lifetime0 + 1)[:-1]
        lifetime_extra = jnp.where(link == -1, lifetime0 + 1, 0)
        return jnp.concatenate([lifetime, lifetime_extra])

    indicator_for_yx0 = (selected0 == 1) & (link == -1)  # [n_samples, n_source]
    indicator_for_yx1 = selected1 == 1  # [n_samples, n_target]
    indicator_all = np.concatenate(
        [indicator_for_yx1, indicator_for_yx0], axis=-1
    )  # [n_sample, n_source+n+target]
    is_not_empty = indicator_all.any(axis=0)  # [n_source+n_target]

    yx_all = np.concatenate([yx1, yx0])  # [n_sourc+n_target, 2]
    id_all = np.concatenate([id1, id0])
    lifetime1 = np.asarray(jax.vmap(_update_lifetime)(lifetime0, link, div))
    # print(lifetime1.shape)

    history["detections"].append(
        dict(
            yx=yx_all[is_not_empty],
            id=id_all[is_not_empty],
        )
    )
    history["samples"][-1].update(
        dict(
            selected=indicator_all[:, is_not_empty].astype(int),
            lifetime=lifetime1[:, is_not_empty],
        )
    )

    return history


def get_tracking_weights(yx0, yx1, gamma):
    delta = ((yx1[None, :, :] - yx0[:, None, :]) ** 2).sum(
        axis=-1
    )  # [n_source, n_target]
    wts = jnp.exp(-delta / 2 * gamma * gamma) * gamma * gamma / 0.399
    return wts


def compute_div_p(lifetime, avg, scale, limit):
    p0 = (1 / scale) ** avg
    return p0 * (scale**lifetime) + limit


@dataclasses.dataclass
class HyperParams:
    gamma: float = 0.1
    div_avg: float = 30.0
    div_limit: float = 0.1
    div_scale: float = 0.9
    w_miss: float = 0.1
    logit_scale: float = 1.0
    logit_offset: float = 0.0
    miss_weight: float = 1.0
    miss_logit: float = -0.5
    n_sub_sample: int = 256
    p_death: float = 0.02


def track_to_next_frame(key, history, nextframe, hyper_params):
    """
    Args:
        key: rng
        nextframe: dict
          yx: [n_target, 2], yx locations
          id: [n_target] int
          logit: [n_target], logit of each yx1 location
        history: dict['yx': [array,...], 'samples': [{'selected':array, 'history':array, 'history_div':array, 'lifetime':array},...]]
            yx: [n_source, 2], yx locations from frame 0
            selected: [n_samples, n_source]: indicators of source locations that are being tracked
            lifetime: [n_samples, n_source] how long has each sample been without division
        haper_params: HyperParam dataclass
    """
    yx0 = history["detections"][-1]["yx"]
    selected0 = history["samples"][-1]["selected"]
    lifetime = history["samples"][-1]["lifetime"]

    yx1 = np.asarray(nextframe["yx"])
    id1 = np.asarray(nextframe["id"])
    logit = np.asarray(nextframe["logit"])
    target_logit = logit * hyper_params.logit_scale + hyper_params.logit_offset
    miss_logit = (
        hyper_params.miss_logit * hyper_params.logit_scale + hyper_params.logit_offset
    )

    n_sample, _ = selected0.shape
    wts = get_tracking_weights(yx0, yx1, hyper_params.gamma)  # [n_source, n_target]
    w_miss = (1 - wts.sum(axis=1)) * np.exp(miss_logit) * hyper_params.miss_weight

    wts = wts * np.exp(target_logit)

    w_nd = compute_div_p(
        lifetime,
        hyper_params.div_avg,
        hyper_params.div_scale,
        hyper_params.div_limit,
    )

    key, ckey = jax.random.split(key)
    new_history, selected1 = sample_step(
        ckey,
        selected0,
        wts,
        w_miss,
        w_nd,
        n_sub_sample=hyper_params.n_sub_sample,
    )

    # random remove missing cell
    is_missing = selected0 & (new_history[:, :, 0] == -1)
    if hyper_params.p_death > 0:
        key, ckey = jax.random.split(key)
        is_dead = jax.random.uniform(ckey, is_missing.shape) < hyper_params.p_death
        new_history[is_dead & is_missing, 0] = -2

    # save new tracks
    history["detections"][-1].update(
        dict(
            yx_next=yx1,
            id_next=id1,
            logit_next=logit,
        )
    )
    history["samples"][-1].update(
        dict(
            next=new_history,
        )
    )
    history["samples"].append(
        dict(
            selected=np.asarray(selected1),
        )
    )

    # resample all chains
    target_logit = (target_logit * selected1).sum(axis=-1)
    target_logit += is_missing.sum(axis=1) * hyper_params.miss_logit
    weights = jax.nn.softmax(target_logit)
    rs = np.asarray(jax.random.choice(key, n_sample, [n_sample], p=weights))
    history["samples"] = jax.tree_map(lambda v: v[rs], history["samples"])

    history = post_process_step(history)

    return history



# original implementation is from treex https://github.com/cgarciae/treex
# which is itself based on Tensorflow Keras:
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/losses.py#L44-L201

import typing as tp
from abc import abstractmethod
from enum import Enum

import jax.numpy as jnp
import numpy as np
from flax import struct

from ..utils import _get_name

IndexLike = tp.Union[str, int, tp.Sequence[tp.Union[str, int]]]
PathLike = tp.Tuple[IndexLike, ...]
ScalarLike = tp.Union[float, np.ndarray, jnp.ndarray]


class Reduction(Enum):
    """
    Types of loss reduction.

    Contains the following values:
    * `NONE`: Weighted losses with one dimension reduced (axis=-1, or axis
        specified by loss function). When this reduction type used with built-in
        Keras training loops like `fit`/`evaluate`, the unreduced vector loss is
        passed to the optimizer but the reported loss will be a scalar value.
    * `SUM`: Scalar sum of weighted losses.
    * `SUM_OVER_BATCH_SIZE`: Scalar `SUM` divided by number of elements in losses.
    """

    # AUTO = "auto"
    NONE = "none"
    SUM = "sum"
    SUM_OVER_BATCH_SIZE = "sum_over_batch_size"

    @classmethod
    def all(cls):
        return (
            # cls.AUTO,
            cls.NONE,
            cls.SUM,
            cls.SUM_OVER_BATCH_SIZE,
        )

    @classmethod
    def validate(cls, key):
        if key not in cls.all():
            raise ValueError("Invalid Reduction Key %s." % key)


class Loss:
    """
    Loss base class.

    To be implemented by subclasses:

    * `call()`: Contains the logic for loss calculation.

    Example subclass implementation:

    ```python
    class MeanSquaredError(Loss):
        def call(self, target, preds):
            return jnp.mean(jnp.square(preds - target), axis=-1)
    ```

    Please see the [Modules, Losses, and Metrics Guide]
    (https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#losses) for more
    details on this.
    """

    def __init__(
        self,
        reduction: tp.Optional[Reduction] = None,
        weight: tp.Optional[ScalarLike] = None,
        on: tp.Optional[IndexLike] = None,
        name: tp.Optional[str] = None,
    ):
        """
        Initializes `Loss` class.

        Arguments:
            reduction: (Optional) Type of `tx.losses.Reduction` to apply to
                loss. Default value is `SUM_OVER_BATCH_SIZE`. For almost all cases
                this defaults to `SUM_OVER_BATCH_SIZE`.
            weight: Optional weight contribution for the total loss. Defaults to `1`.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `target` and `preds`
                arguments before passing them to `call`. For example if `on = "a"` then
                `target = target["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `target = target["a"][0]["b"]`, same for `preds`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
            name: Optional name for the instance, if not provided lower snake_case version
                of the name of the class is used instead.
        """
        self.name = name if name is not None else _get_name(self)
        self.weight = (
            jnp.asarray(weight, dtype=jnp.float32)
            if weight is not None
            else jnp.array(1.0, dtype=jnp.float32)
        )
        self._reduction = reduction if reduction is not None else Reduction.NONE
        self._labels_filter = (on,) if isinstance(on, (str, int)) else on
        self._signature_f = self.call

    def __call__(
        self,
        **kwargs,
    ) -> jnp.ndarray:

        if self._labels_filter is not None:
            if "target" in kwargs and kwargs["target"] is not None:
                for index in self._labels_filter:
                    kwargs["target"] = kwargs["target"][index]

            if "preds" in kwargs and kwargs["preds"] is not None:
                for index in self._labels_filter:
                    kwargs["preds"] = kwargs["preds"][index]

        sample_weight: tp.Optional[jnp.ndarray] = kwargs.pop("sample_weight", None)

        values = self.call(**kwargs)

        return reduce_loss(values, sample_weight, self.weight, self._reduction)

    def __hash__(self):
        return hash(self.name)

    @abstractmethod
    def call(self, *args, **kwargs) -> jnp.ndarray:
        ...


def reduce_loss(
    values: jnp.ndarray, sample_weight: tp.Optional[jnp.ndarray], weight, reduction
) -> jnp.ndarray:

    values = jnp.asarray(values)

    if sample_weight is not None:
        # expand `sample_weight` dimensions until it has the same rank as `values`
        while sample_weight.ndim < values.ndim:
            sample_weight = sample_weight[..., jnp.newaxis]

        values *= sample_weight

    if reduction == Reduction.NONE:
        loss = values
    elif reduction == Reduction.SUM:
        loss = jnp.sum(values)
    elif reduction == Reduction.SUM_OVER_BATCH_SIZE:
        loss = jnp.sum(values) / jnp.prod(jnp.array(values.shape))
    else:
        raise ValueError(f"Invalid reduction '{reduction}'")

    return loss * weight


class LossLog(struct.PyTreeNode):
    loss_fn: Loss = struct.field(pytree_node=False)
    weight: float = 1.0
    cnt: jnp.ndarray = jnp.array(0.0)
    sum: jnp.ndarray = jnp.array(0.0)

    def update(self, **kwargs):
        loss = self.loss_fn(**kwargs) * self.weight
        new_log = self.replace(cnt=self.cnt + 1, sum=self.sum + loss)
        return loss, new_log

    def compute(self):
        return self.sum / self.cnt



from functools import partial

import jax

from ..train.loss import Loss

jnp = jax.numpy
EPS = jnp.finfo("float32").eps


def instance_overlap_losses(
    instances, instance_logit, yc, xc, mask, seg, ignore_seg_loss=False
):
    """
    Args:
        instances: [N, H, W]
        instances_logit: [N, H, W]
        yc: [N, H, W]
        xc: [N, H, W]
        mask: [N, 1, 1]
        seg: [img_height, img_width]
        ignore_seg_loss: bool
    return:
        loss:
    """
    n_instances = jnp.count_nonzero(mask)

    patch_size = instances.shape[-1]
    padding_size = patch_size // 2 + 2
    yc += padding_size
    xc += padding_size

    seg = jnp.pad(seg, padding_size).astype(instances.dtype)
    if not ignore_seg_loss:
        loss = (
            (seg.sum() - instances.sum())
            / patch_size
            / patch_size
            / (n_instances + EPS)
        )
        log_yi_sum = (1.0 - seg) * (-2.0)
    else:
        loss = 0.0
        log_yi_sum = jnp.zeros_like(seg)

    log_yi = -jnp.log(1 + jnp.exp(instance_logit))
    log_yi_sum = log_yi_sum.at[yc, xc].add(log_yi)
    log_yi = log_yi_sum[yc, xc] - log_yi

    loss = loss - (instances * log_yi).mean(axis=(1, 2), keepdims=True).sum(
        where=mask
    ) / (n_instances + EPS)

    return loss


def supervised_segmentation_losses(instances, yc, xc, mask, gt_label):
    """
    Args:
        instances: [N, H, W]
        yc: [N, H, W]
        xc: [N, H, W]
        mask: [N, 1, 1]
        gt_label: [img_height, img_width] labels, bg_label=0
    return:
        loss: based on cross entropy
    """
    n_patches, _, _ = yc.shape

    gt_label = gt_label.astype(int)
    gt_label = jnp.pad(gt_label, [[1, 1], [1, 1]])
    gt_patches = gt_label[yc + 1, xc + 1] == (jnp.arange(n_patches)[:, None, None] + 1)

    p_t = gt_patches * instances + (1 - gt_patches) * (1.0 - instances)
    bce = -jnp.log(jnp.clip(p_t, EPS, 1.0))

    loss = bce.mean(axis=(1, 2), keepdims=True).sum(where=mask) / (mask.sum() + EPS)

    return loss


class InstanceLoss(Loss):
    def call(self, binary_mask: jnp.ndarray, preds: dict, **kwargs):
        if not "training_locations" in preds:
            return 0.0
        return instance_overlap_losses(
            instances=preds["instance_output"],
            instance_logit=preds["instance_logit"],
            yc=preds["instance_yc"],
            xc=preds["instance_xc"],
            mask=preds["instance_mask"],
            seg=binary_mask,
        )


class InstanceOverlapLoss(Loss):
    def call(self, inputs: dict, preds: dict, **kwargs):
        if not "training_locations" in preds:
            return 0.0
        segs = jnp.ones(inputs["image"].shape[:-1])
        op = partial(instance_overlap_losses, ignore_seg_loss=True)
        return op(
            instances=preds["instance_output"],
            instance_logit=preds["instance_logit"],
            yc=preds["instance_yc"],
            xc=preds["instance_xc"],
            mask=preds["instance_mask"],
            seg=segs,
        )


class SupervisedInstanceLoss(Loss):
    def call(self, mask_labels: jnp.ndarray, preds: dict, **kwargs):
        if not "training_locations" in preds:
            return 0.0

        return supervised_segmentation_losses(
            instances=preds["instance_output"],
            yc=preds["instance_yc"],
            xc=preds["instance_xc"],
            mask=preds["instance_mask"],
            gt_label=mask_labels,
        )



import dataclasses
import re
import typing as tp

import numpy as np
from flax import struct

from .ops import bboxes_of_patches

InputLike = tp.Union[tp.Any, tp.Tuple[tp.Any, ...], tp.Dict[str, tp.Any], "Inputs"]


def _unique_name(
    names: tp.Set[str],
    name: str,
):

    if name in names:

        match = re.match(r"(.*?)(\d*)$", name)
        assert match is not None

        name = match[1]
        num_part = match[2]

        i = int(num_part) if num_part else 2
        str_template = f"{{name}}{{i:0{len(num_part)}}}"

        while str_template.format(name=name, i=i) in names:
            i += 1

        name = str_template.format(name=name, i=i)

    names.add(name)
    return name


def _unique_names(
    names: tp.Iterable[str],
    *,
    existing_names: tp.Optional[tp.Set[str]] = None,
) -> tp.Iterable[str]:
    if existing_names is None:
        existing_names = set()

    for name in names:
        yield _unique_name(existing_names, name)


def _lower_snake_case(s: str) -> str:
    s = re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()
    parts = s.split("_")
    output_parts = []

    for i in range(len(parts)):
        if i == 0 or len(parts[i - 1]) > 1:
            output_parts.append(parts[i])
        else:
            output_parts[-1] += parts[i]

    return "_".join(output_parts)


def _get_name(obj) -> str:
    if hasattr(obj, "name") and obj.name:
        return obj.name
    elif hasattr(obj, "__name__") and obj.__name__:
        return _lower_snake_case(obj.__name__)
    elif hasattr(obj, "__class__") and obj.__class__.__name__:
        return _lower_snake_case(obj.__class__.__name__)
    else:
        raise ValueError(f"Could not get name for: {obj}")


class Inputs(struct.PyTreeNode):
    args: tp.Tuple[tp.Any, ...] = ()
    kwargs: tp.Dict[str, tp.Any] = dataclasses.field(default_factory=dict)

    def update(self, *args, **kwargs):
        tmp = self.kwargs.copy()
        tmp.update(kwargs)
        new_inputs = self.replace(args=self.args + args, kwargs=tmp)
        return new_inputs

    @classmethod
    def from_value(cls, value: InputLike) -> "Inputs":
        if isinstance(value, cls):
            return value
        elif isinstance(value, tuple):
            return cls(args=value)
        elif isinstance(value, dict):
            return cls(kwargs=value)
        else:
            return cls(args=(value,))


def _to_str(p):
    return "".join(p.astype(int).reshape(-1).astype(str).tolist())


def format_predictions(pred, mask=None, threshold=0.5):
    """
    Args:
        pred: model output without batch dim
            instance_output: [n, patch_size, patch_size] float
            instance_yx, instance_xc: [n, patch_size, patch_size]
            pred_scores: [n]
            instance_mask: [n, 1, 1]
        mask: optional mask selecting cells
        threshold: float
    Returns:
        bboxes: [n, 4] int64
        encodings: [n] string
    """
    patches = np.asarray(pred["instance_output"]) > threshold
    yc = np.asarray(pred["instance_yc"])
    xc = np.asarray(pred["instance_xc"])
    scores = np.asarray(pred["pred_scores"])
    bboxes = np.asarray(bboxes_of_patches(pred))

    is_valid = pred["instance_mask"].squeeze(axis=(-1, -2))
    is_valid & patches.any(axis=(1, 2))  # no empty patches
    if mask is not None:
        is_valid &= mask

    patches = patches[is_valid]
    yc = yc[is_valid]
    xc = xc[is_valid]
    scores = scores[is_valid]

    encodings = []
    for (r0, c0, r1, c1), y0, x0, p in zip(bboxes, yc[:, 0, 0], xc[:, 0, 0], patches):
        roi = p[r0 - y0 : r1 - y0, c0 - x0 : c1 - x0]
        encodings.append(_to_str(roi))

    n_pixels = np.count_nonzero(patches, axis=(1, 2))
    yx = np.stack(
        [
            (patches * yc).sum(axis=(1, 2)) / n_pixels,
            (patches * xc).sum(axis=(1, 2)) / n_pixels,
        ],
        axis=-1,
    )  # centroid

    return bboxes, encodings, scores, yx



from typing import List, Sequence, Union

import flax.linen as nn
import jax
import jax.numpy as jnp

from .common import *

""" Implements the convnext encoder. Described in https://arxiv.org/abs/2201.03545
Original implementation: https://github.com/facebookresearch/ConvNeXt
"""


class _Block(nn.Module):
    """ConvNeXt Block.
    Args:
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    drop_rate: int = 0.0
    layer_scale_init_value: float = 1e-6
    kernel_size: int = 7

    @nn.compact
    def __call__(self, x: jnp.ndarray, *, training=None) -> jnp.ndarray:
        dim = x.shape[-1]
        ks = self.kernel_size
        scale = self.layer_scale_init_value

        shortcut = x

        x = nn.Conv(dim, (ks, ks), feature_group_count=dim)(x)
        x = nn.LayerNorm(epsilon=1e-6)(x)
        x = nn.Dense(dim * 4)(x)
        x = jax.nn.gelu(x)
        x = nn.Dense(dim)(x)

        if scale > 0:
            gamma = self.param(
                "gamma", lambda rng, shape: scale * jnp.ones(shape), (x.shape[-1])
            )
            x = x * gamma

        deterministic = training is None or not training
        x = DropPath(self.drop_rate)(x, deterministic=deterministic)

        x = x + shortcut

        return x


class ConvNeXt(nn.Module):
    """ConvNeXt
    Args:
        patch_size: for stem default 4
        depths (tuple(int)): Number of blocks at each stage. Default: [3, 3, 9, 3]
        dims (int): Feature dimension at each stage. Default: [96, 192, 384, 768]
        drop_path_rate (float): Stochastic depth rate. Default: 0.
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
        out_channels (int): FPN output channels. Default: 256
    """

    patch_size: int = 4
    depths: Sequence[int] = (3, 3, 9, 3)
    dims: Sequence[int] = (96, 192, 384, 768)
    drop_path_rate: float = 0.0
    layer_scale_init_value: float = 1e-6
    out_channels: int = 256

    @nn.compact
    def __call__(self, x: jnp.ndarray, *, training: bool = None) -> jnp.ndarray:
        dp_rate = 0
        outputs = []
        for k in range(len(self.depths)):
            if k == 0:
                ps = self.patch_size
                x = nn.Conv(self.dims[k], (ps, ps), strides=(ps, ps))(x)
                x = nn.LayerNorm(epsilon=1e-6)(x)
            else:
                x = nn.LayerNorm(epsilon=1e-6)(x)
                x = nn.Conv(self.dims[k], (2, 2), strides=(2, 2))(x)

            for _ in range(self.depths[k]):
                x = _Block(dp_rate, self.layer_scale_init_value)(x, training=training)
                dp_rate += self.drop_path_rate / (sum(self.depths) - 1)

            outputs.append(x)

        decoder_out = FPN(self.out_channels)(outputs)
        keys = [str(k + 1 if self.patch_size == 2 else k + 2) for k in range(4)]
        encoder_out = dict(zip(keys, outputs))
        decoder_out = dict(zip(keys, decoder_out))

        return encoder_out, decoder_out


# utility funcs

model_urls = {
    "convnext_tiny_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_1k_224_ema.pth",
    "convnext_small_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_1k_224_ema.pth",
    "convnext_base_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_1k_224_ema.pth",
    "convnext_large_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_1k_224_ema.pth",
    "convnext_tiny_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_22k_224.pth",
    "convnext_small_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_22k_224.pth",
    "convnext_base_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_22k_224.pth",
    "convnext_large_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_22k_224.pth",
    "convnext_xlarge_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_xlarge_22k_224.pth",
}

# def load_weight(jax_model, url):
#     ''' Load pretrained convnext models
#     specs for pretrained models are:
#         tiny: dims=(96, 192, 384, 768), depths=(3,3,9,3)
#         small: dims=(96, 192, 384, 768), depths=(3,3,27,3)
#         base: dims=(128, 256, 512, 1024), depths=(3,3,27,3)
#         large: dims=(192, 384, 768, 1536), depths=(3,3,27,3)
#         X-large: dims=(256, 512, 1024, 2048), depths=(3,3,27,3)
#     '''
#     import torch

#     def t(m, k, new_value):
#         new_value = jnp.array(new_value)
#         assert vars(m)[k].shape == new_value.shape
#         vars(m)[k] = new_value

#     checkpoint = torch.hub.load_state_dict_from_url(url=url, map_location="cpu")
#     params = checkpoint['model']

#     cnt = 1
#     for i in range(4):
#         for k in range(jax_model.depths[i]):
#             block = vars(jax_model)[f'block{cnt}' if cnt > 1 else 'block']
#             t(block, 'gamma', params[f'stages.{i}.{k}.gamma'])
#             t(block.conv, 'bias', params[f'stages.{i}.{k}.dwconv.bias'])
#             t(block.conv, 'kernel', params[f'stages.{i}.{k}.dwconv.weight'].permute(2,3,1,0))
#             t(block.layer_norm, 'bias', params[f'stages.{i}.{k}.norm.bias'])
#             t(block.layer_norm, 'scale', params[f'stages.{i}.{k}.norm.weight'])
#             t(block.linear, 'bias', params[f'stages.{i}.{k}.pwconv1.bias'])
#             t(block.linear, 'kernel', params[f'stages.{i}.{k}.pwconv1.weight'].transpose(0,1))
#             t(block.linear2, 'bias', params[f'stages.{i}.{k}.pwconv2.bias'])
#             t(block.linear2, 'kernel', params[f'stages.{i}.{k}.pwconv2.weight'].transpose(0,1))

#             cnt += 1

#     norm = jax_model.layer_norm
#     t(norm, 'bias', params['downsample_layers.0.1.bias'])
#     t(norm, 'scale', params['downsample_layers.0.1.weight'])

#     conv = jax_model.conv
#     t(conv, 'bias', params['downsample_layers.0.0.bias'])
#     t(conv, 'kernel', params['downsample_layers.0.0.weight'].permute(2,3,1,0))

#     for i in range(1,4):
#         norm = vars(jax_model)[f'layer_norm{i+1}']
#         t(norm, 'bias', params[f'downsample_layers.{i}.0.bias'])
#         t(norm, 'scale', params[f'downsample_layers.{i}.0.weight'])
#         conv = vars(jax_model)[f'conv{i+1}']
#         t(conv, 'bias', params[f'downsample_layers.{i}.1.bias'])
#         t(conv, 'kernel', params[f'downsample_layers.{i}.1.weight'].permute(2,3,1,0))

#     return jax_model



import typing as tp
from functools import partial

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from flax.training.train_state import TrainState

from ..utils import Inputs
from .loss import Loss, LossLog


class Eager:
    @classmethod
    def loss_fn(cls, params, state, loss_logs, inputs, labels, rngs):
        inputs_obj = Inputs.from_value(inputs)
        preds = state.apply_fn(
            {"params": params},
            *inputs_obj.args,
            **inputs_obj.kwargs,
            rngs=rngs,
        )

        args = dict(
            inputs=inputs,
            preds=preds,
            **labels,
        )

        losses, loss_logs = zip(*[loss_fn.update(**args) for loss_fn in loss_logs])
        total_loss = sum(losses)

        return total_loss, (state, loss_logs, preds)

    @classmethod
    def init_fn(cls, key, model, inputs, tx):
        inputs_obj = Inputs.from_value(inputs)
        variables = model.init(key, *inputs_obj.args, **inputs_obj.kwargs)
        state = TrainState.create(
            apply_fn=model.apply,
            params=variables["params"],
            tx=tx,
        )
        return state

    @classmethod
    def predict(cls, state, inputs):  # only if model is immutable
        # print("JIT Predict")
        inputs_obj = Inputs.from_value(inputs)
        preds = state.apply_fn(
            {"params": state.params}, *inputs_obj.args, **inputs_obj.kwargs
        )
        return preds

    @classmethod
    def train_step(
        cls,
        state: TrainState,
        loss_logs: tp.Sequence[LossLog],
        inputs: tp.Any,
        labels: tp.Any,
        rngs: tp.Optional[dict],
    ) -> tp.Tuple[TrainState, tp.Sequence[LossLog], tp.Any]:
        # print('JIT train_step')

        if labels is None:
            labels = {}
        elif not isinstance(labels, dict):
            labels = dict(target=labels)

        grads, (state, loss_logs, preds) = jax.grad(cls.loss_fn, has_aux=True)(
            state.params,
            state,
            loss_logs,
            inputs,
            labels,
            rngs,
        )
        state = state.apply_gradients(grads=grads)

        return state, loss_logs, preds


class Core(Eager):
    predict = jax.jit(Eager.predict)


class JIT(Core):
    train_step = jax.jit(Core.train_step)


class _Distributed(Eager):
    @classmethod
    def _train_step(
        cls,
        state: TrainState,
        loss_logs: tp.Sequence[LossLog],
        inputs: tp.Any,
        labels: tp.Any,
        rngs: tp.Optional[dict],
    ) -> tp.Tuple[TrainState, tp.Sequence[LossLog], tp.Any]:
        # print("JITTTTING")
        axis_index = jax.lax.axis_index("mapped")
        rngs = jax.tree_util.tree_map(
            lambda key: jax.random.fold_in(key, axis_index), rngs
        )

        if labels is None:
            labels = {}
        elif not isinstance(labels, dict):
            labels = dict(target=labels)

        grads, (state, loss_logs, preds) = jax.grad(cls.loss_fn, has_aux=True)(
            state.params,
            state,
            loss_logs,
            inputs,
            labels,
            rngs,
        )

        grads = jax.lax.pmean(grads, axis_name="mapped")
        state = state.apply_gradients(grads=grads)

        # aggregate logs
        loss_logs = jax.tree_map(partial(jax.lax.pmean, axis_name="mapped"), loss_logs)

        #         # sync batch statistics
        #         model.map(partial(jax.lax.pmean, axis_name="mapped"), tx.BatchStat, inplace=True)

        return state, loss_logs, preds

    @classmethod
    def init_fn(cls, key, model, inputs, tx):
        inputs = jax.tree_map(lambda v: v[0], inputs)
        return Eager.init_fn(key, model, inputs, tx)


class Distributed(_Distributed):
    train_step = jax.pmap(
        _Distributed._train_step,
        axis_name="mapped",
        in_axes=(None, None, 0, 0, None),
        out_axes=(None, None, 0),
    )

    predict = jax.pmap(
        Eager.predict,
        axis_name="mapped",
        in_axes=(None, 0),
    )


class VMapped(_Distributed):
    train_step = jax.jit(
        jax.vmap(
            _Distributed._train_step,
            axis_name="mapped",
            in_axes=(None, None, 0, 0, None),
            out_axes=(None, None, 0),
        )
    )

    predict = jax.jit(
        jax.vmap(
            Eager.predict,
            axis_name="mapped",
            in_axes=(None, 0),
        )
    )



from functools import partial
from typing import Optional

import jax
import optax

from ..train.loss import Loss

jnp = jax.numpy

EPS = jnp.finfo("float32").eps


def _binary_focal_crossentropy(pred, gt, gamma=2.0):
    p_t = gt * pred + (1 - gt) * (1.0 - pred)
    focal_factor = (1.0 - p_t) ** gamma

    bce = -jnp.log(jnp.clip(p_t, EPS, 1.0))

    return focal_factor * bce


class DetectionLoss(Loss):
    def __init__(self, gamma=2.0, **kwargs):
        super().__init__(**kwargs)
        self.gamma = gamma

    def call(self, preds: dict, **kwargs) -> jnp.ndarray:
        def _inner(scores, gt_scores):
            score_loss = 0.0
            cnt = 1e-8
            for k in scores:
                score_loss += _binary_focal_crossentropy(
                    scores[k],
                    gt_scores[k],
                    self.gamma,
                ).sum()
                cnt += preds["lpn_scores"][k].size
            return score_loss / cnt

        return _inner(preds["lpn_scores"], preds["lpn_gt_scores"])


class LocalizationLoss(Loss):
    def __init__(self, delta=1.0, **kwargs):
        super().__init__(**kwargs)
        self.delta = delta

    def call(self, preds: dict, **kwrags) -> jnp.ndarray:
        def _inner(regrs, gt_regrs, gt_scores):
            regr_loss = 0.0
            cnt = 1e-8
            for k in regrs:
                # h = optax.l2_loss(regrs[k], gt_regrs[k])
                h = optax.huber_loss(regrs[k], gt_regrs[k], delta=self.delta).mean(
                    axis=-1, keepdims=True
                )
                mask = gt_scores[k] > 0
                regr_loss += jnp.sum(h, where=mask)
                cnt += jnp.count_nonzero(mask)
            return regr_loss / cnt

        return _inner(
            preds["lpn_regressions"],
            preds["lpn_gt_regressions"],
            preds["lpn_gt_scores"],
        )


# class LocalizationLossAlt(Loss):
#     def __init__(self, delta=1.0, sigma=2.0, **kwargs):
#         super().__init__(**kwargs)
#         self.delta = delta
#         self.sigma_sq = sigma * sigma

#     def call(self, preds: dict, **kwargs) -> jnp.ndarray:
#         def _inner(regrs, gt_regrs, gt_scores):
#             regr_loss = 0.0
#             cnt = 1e-8
#             for k in regrs:
#                 # h = optax.l2_loss(regrs[k], gt_regrs[k])
#                 h = optax.huber_loss(regrs[k], gt_regrs[k], delta=self.delta).mean(
#                     axis=-1
#                 )
#                 w = jnp.exp(-(gt_regrs[k] ** 2).sum(axis=-1) / self.sigma_sq)
#                 regr_loss += jnp.sum(h * w)
#                 cnt += w.sum()
#             return regr_loss / cnt

#         return jax.vmap(_inner)(
#             preds["lpn_regressions"],
#             preds["lpn_gt_regressions"],
#             preds["lpn_gt_scores"],
#         )


class LPNLoss(Loss):
    def __init__(self, delta=1.0, gamma=2.0, **kwargs):
        super().__init__(**kwargs)
        self.det_loss = DetectionLoss(gamma=gamma)
        self.loc_loss = LocalizationLoss(delta=delta)

    def call(self, preds: dict, **kwargs):
        return self.det_loss.call(preds=preds) + self.loc_loss(preds=preds)



import jax
import jax.numpy as jnp
import numpy as np


def box_area(box):
    """Computes area of boxes.
    Args:
      box: a float Tensor with [..., N, 4].

    Returns:
      a float Tensor with [..., N]
    """
    if isinstance(box, jnp.ndarray):
        split = jnp.split
    else:
        split = np.split

    y_min, x_min, y_max, x_max = split(box, 4, axis=-1)
    return ((y_max - y_min) * (x_max - x_min)).squeeze(axis=-1)


def box_intersection(gt_boxes, boxes):
    """Compute pairwise intersection areas between boxes.

    Args:
      gt_boxes: [..., N, 4]
      boxes: [..., M, 4]

    Returns:
      a float Tensor with shape [..., N, M] representing pairwise intersections.
    """
    if isinstance(gt_boxes, jnp.ndarray) & isinstance(boxes, jnp.ndarray):
        split = jnp.split
        minimum = jnp.minimum
        maximum = jnp.maximum
    else:
        split = np.split
        minimum = np.minimum
        maximum = np.maximum

    y_min1, x_min1, y_max1, x_max1 = split(gt_boxes, 4, axis=-1)
    y_min2, x_min2, y_max2, x_max2 = split(boxes, 4, axis=-1)

    # [N, M] or [B, N, M]
    y_min_max = minimum(y_max1, y_max2.swapaxes(-1, -2))
    y_max_min = maximum(y_min1, y_min2.swapaxes(-1, -2))
    x_min_max = minimum(x_max1, x_max2.swapaxes(-1, -2))
    x_max_min = maximum(x_min1, x_min2.swapaxes(-1, -2))

    intersect_heights = y_min_max - y_max_min
    intersect_widths = x_min_max - x_max_min
    intersect_heights = maximum(0, intersect_heights)
    intersect_widths = maximum(0, intersect_widths)

    return intersect_heights * intersect_widths


def box_iou_similarity(gt_boxes, boxes):
    """Computes pairwise intersection-over-union between box collections.

    Args:
      gt_boxes: a float Tensor with [..., N, 4].
      boxes: a float Tensor with [..., M, 4].

    Returns:
      a Tensor with shape [..., N, M] representing pairwise iou scores.
    """
    intersections = box_intersection(gt_boxes, boxes)
    gt_boxes_areas = box_area(gt_boxes)
    boxes_areas = box_area(boxes)
    unions = gt_boxes_areas[..., None] + boxes_areas[..., None, :] - intersections

    ious = intersections / (unions + 1e-6)

    return ious



from typing import Optional, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp


class ChannelAttention(nn.Module):
    squeeze_factor: int = 16

    @nn.compact
    def __call__(
        self,
        x: jnp.ndarray,
    ) -> jnp.ndarray:

        orig_shape = x.shape

        x = x.reshape(-1, orig_shape[-1])
        x_backup = x

        x = jnp.stack([x.max(axis=0), x.mean(axis=0)])

        se_channels = orig_shape[-1] // self.squeeze_factor
        x = nn.Dense(se_channels)(x)
        x = jax.nn.relu(x)
        x = x[0] + x[1]

        x = nn.Dense(orig_shape[-1])(x)
        x = jax.nn.sigmoid(x)

        x = x_backup * x
        x = x.reshape(orig_shape)

        return x


class SpatialAttention(nn.Module):
    filter_size: int = 7

    @nn.compact
    def __call__(
        self,
        x: jnp.ndarray,
    ) -> jnp.ndarray:

        y = jnp.stack([x.max(axis=-1), x.mean(axis=-1)], axis=-1)
        y = nn.Conv(1, [self.filter_size, self.filter_size])(y)
        y = jax.nn.sigmoid(y)

        y = x * y

        return y


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks)."""

    rate: float

    @nn.module.compact
    def __call__(self, inputs, deterministic: Optional[bool] = True):
        if self.rate == 0.0:
            return inputs
        keep_prob = 1.0 - self.rate
        if deterministic:
            return inputs
        else:
            rng = self.make_rng("droppath")
            binary_factor = jnp.floor(
                keep_prob + jax.random.uniform(rng, dtype=inputs.dtype)
            )
            output = inputs / keep_prob * binary_factor
            return output


class FPN(nn.Module):
    out_channels: int = 256

    @nn.compact
    def __call__(self, inputs: Sequence[jnp.ndarray]) -> Sequence[jnp.ndarray]:
        out_channels = self.out_channels

        outputs = [jax.nn.relu(nn.Dense(out_channels)(x)) for x in inputs]

        for k in range(len(outputs) - 1, 0, -1):
            x = jax.image.resize(outputs[k], outputs[k - 1].shape, "nearest")
            x += outputs[k - 1]
            x = nn.Conv(out_channels, (3, 3))(x)
            x = jax.nn.relu(x)
            outputs[k - 1] = x

        return outputs



import jax

from .boxes import box_iou_similarity
from .patches import bboxes_of_patches

jnp = jax.numpy


class MaskIndices:
    _row_lengths = None
    _row_starts = None

    def __init__(self, data):
        self.data = data

    def _compute_row_lengths(self):
        diff_ids = self.data[1:, -1] - self.data[:-1, -1]
        jump_locs = jnp.argwhere(diff_ids)
        jump_locs = jump_locs.repeat(diff_ids[jump_locs]) + 1
        jump_locs = jnp.concatenate(
            [jnp.array([0]), jump_locs, jnp.array([self.data.shape[0]])]
        )

        self._row_starts = jump_locs[:-1]
        self._row_lengths = jump_locs[1:] - jump_locs[:-1]

    @property
    def row_lengths(self):
        if self._row_lengths is None:
            self._compute_row_lengths()
        return self._row_lengths

    @property
    def row_starts(self):
        if self._row_starts is None:
            self._compute_row_lengths()
        return self._row_starts

    @property
    def n_masks(self):
        return len(self.row_lengths)

    def gather(self, indices):
        sel_starts = self.row_starts[indices]
        sel_repeats = self.row_lengths[indices]
        new_data = jnp.concatenate(
            [self.data[s : s + l, :2] for s, l in zip(sel_starts, sel_repeats)]
        )
        new_ids = jnp.arange(sel_starts.size).repeat(sel_repeats)

        new_masks = MaskIndices(jnp.append(new_data, new_ids[:, None], axis=-1))
        new_masks._row_lengths = sel_repeats
        new_masks._row_starts = jnp.insert(sel_repeats[:-1].cumsum(), 0, 0)

        return new_masks

    def to_patches(self, y0x0, patch_size: int):
        y0x0_repeated = y0x0.repeat(self.row_lengths, axis=0)
        shifted_data = self.data.at[:, :2].add(-y0x0_repeated)
        patches = jnp.zeros([self.n_masks, patch_size, patch_size], dtype=int)
        patches = patches.at[
            shifted_data[:, 2], shifted_data[:, 0], shifted_data[:, 0]
        ].set(1)

        return patches

    def bboxes(self):
        min_y, min_x = jax.ops.segment_min(
            self.data[:, :2], self.data[:, -1]
        ).transpose()
        max_y, max_x = (
            jax.ops.segment_max(self.data[:, :2], self.data[:, -1]).transpose() + 1
        )
        bboxes = jnp.stack([min_y, min_x, max_y, max_x], axis=-1)

        bboxes = jnp.where(self.row_lengths[:, None] > 0, bboxes, -1)

        return bboxes


def mask_iou_similarity(gt_masks, pred_masks, gt_boxes, pred_boxes, patch_size):
    """Compute mask_iou matrix.
    The method is to compute box iou matrix first and update only non-zero
    entries with the accurate mask_iou

    Args:
        gt_masks: [N, 3]  ground truth mask indices data. format: [yy, xx, mask_id]
        pred_masks: [M, 3] prediction mask indices data. format: [yy, xx, mask_id]
        gt_boxes: [N, 4]  y0x0y1x1 format
        pred_boxes: [M, 4] y0x0y1x1 format
        patch_size: int constant
    Outputs:
        similarity_matrix: [n_pred_instances, n_gt_instances]
    """
    gt_masks = MaskIndices(gt_masks)
    pred_masks = MaskIndices(pred_masks)

    if gt_boxes is None:
        gt_boxes = gt_masks.bboxes()
    if pred_boxes is None:
        pred_boxes = pred_masks.bboxes()

    sm = box_iou_similarity(pred_boxes, gt_boxes)

    if (sm > 0).any():
        pred_mask_ids, gt_mask_ids = jnp.argwhere(sm > 0).transpose()

        sub_gt_masks = gt_masks.gather(gt_mask_ids)
        sub_pred_masks = pred_masks.gather(pred_mask_ids)

        y0x0 = pred_boxes[pred_mask_ids, :2]
        sub_gt_patches = sub_gt_masks.to_patches(y0x0, patch_size)
        sub_pred_patches = sub_pred_masks.to_patches(y0x0, patch_size)

        area = sub_gt_masks.row_lengths + sub_pred_masks.row_lengths
        intersects = (sub_gt_patches * sub_pred_patches).sum(axis=(1, 2))
        iou = intersects / (area - intersects + 1e-8)

        sm = sm.at[pred_mask_ids, gt_mask_ids].set(iou)

    return sm


def has_box_intersection(gt_box, pred_box):
    gt_box = gt_box.astype(jnp.float32)
    pred_box = pred_box.astype(jnp.float32)

    y_min1, x_min1, y_max1, x_max1 = jnp.split(gt_box, 4, -1)
    y_min2, x_min2, y_max2, x_max2 = jnp.split(pred_box, 4, -1)

    y_min_max = jnp.minimum(y_max1, jnp.transpose(y_max2))
    y_max_min = jnp.maximum(y_min1, jnp.transpose(y_min2))
    x_min_max = jnp.minimum(x_max1, jnp.transpose(x_max2))
    x_max_min = jnp.maximum(x_min1, jnp.transpose(x_min2))

    intersect_heights = y_min_max - y_max_min
    intersect_widths = x_min_max - x_max_min

    return (intersect_heights > 0) & (intersect_widths > 0)


def mask_intersects(gt_label, gt_box, pred):
    pred_box = bboxes_of_patches(pred)
    its_table = has_box_intersection(gt_box, pred_box)
    gt_ids, pred_ids = jnp.moveaxis(jnp.argwhere(its_table), -1, 0)

    pred_patches = pred["instance_mask"][pred_ids]
    pred_patches = jnp.squeeze(pred_patches, -1)
    pred_patches = pred_patches >= 0.5

    ycs = pred["instance_yc"][pred_ids]
    xcs = pred["instance_xc"][pred_ids]
    gt_patches = gt_label[ycs, xcs] == (gt_ids + 1)[:, None, None]

    its = jnp.count_nonzero(pred_patches & gt_patches, axis=(1, 2))

    intersects = jnp.zeros_like(its_table, dtype=jnp.float32)
    intersects = intersects.at[gt_ids, pred_ids].set(its)

    return intersects


def mask_ious(gt_label, gt_box, pred):
    """Compute mask_iou matrix
    Args:
        gt_label: [H,W]
        gt_box: [N, 4]
        pred: dict. model output without batch dimension.
    Outputs:
        similarity_matrix: Tensor (float32) [n_pred_masks, n_gt_masks]
    """
    n_labels = gt_label.max()
    gt_areas = jnp.count_nonzero(
        gt_label == jnp.arange(1, n_labels + 1)[:, None, None], axis=(1, 2)
    )
    pred_areas = jnp.count_nonzero(pred["instance_mask"] >= 0.5, axis=(1, 2))
    sum_areas = gt_areas[:, None] + pred_areas
    intersects = mask_intersects(gt_label, gt_box, pred)
    ious = intersects / (sum_areas - intersects + 1e-8)

    return jnp.transpose(ious)
